import numpy as np
import pytest

import archimedes as arc


class TestQPSolve:
    def test_casadi_example(self):
        def f(x):
            return x[0] ** 2 + x[1] ** 2

        def g(x):
            return x[0] + x[1] - 10

        x0 = np.array([0.0, 0.0])
        sol = arc.qpsol(f, g, x0, lba=0)

        # Check against the values from the CasADi example
        assert np.allclose(sol.x, [5.0, 5.0], atol=1e-2)
        assert np.allclose(sol.lam_a, -10, atol=1e-2)

        # Test with warm starting
        sol2 = arc.qpsol(f, g, x0, lba=0, lam_a0=sol.lam_a)
        assert np.allclose(sol2.x, sol.x, atol=1e-2)
        assert np.allclose(sol2.lam_a, sol.lam_a, atol=1e-2)

    def test_with_args(self):
        def f(x, a):
            return a * x[0] ** 2 + x[1] ** 2

        def g(x, a):
            return x[0] + x[1] - 10

        x0 = np.array([0.0, 0.0])
        sol = arc.qpsol(f, g, x0, lba=0, args=(1.0,))

        # Check against the values from the CasADi example
        assert np.allclose(sol.x, [5.0, 5.0], atol=1e-2)
        assert np.allclose(sol.lam_a, -10, atol=1e-2)

        # Test with no bounds
        sol = arc.qpsol(f, g, x0, args=(1.0,))
        assert np.allclose(sol.x, [0, 0], atol=1e-2)
        assert np.allclose(sol.lam_a, 0, atol=1e-2)
        # print(sol)

        # # Test warm starting
        # sol2 = qpsol(
        #     f,
        #     g,
        #     sol.x,
        #     lba=0,
        #     lam_x0=sol.lam_x,
        #     lam_a0=sol.lam_a,
        #     check_termination=5,
        #     verbose=True,
        # )

    def test_error_handling(self):
        # Test inconsistent signatures
        def f(x):
            return x[0] ** 2 + x[1] ** 2

        def g(x, y):
            return x[0] + x[1] - 10

        with pytest.raises(ValueError, match=r".*same number of arguments.*"):
            arc.qpsol(f, g, x0=[0.0, 0.0], lba=0)

        # Test inconsistent static arguments
        def f(a, x):
            return a * (x[0] ** 2 + x[1] ** 2)

        def g(a, x):
            return a * (x[0] + x[1]) - 10

        f = arc.compile(f, static_argnames=["a"])
        g = arc.compile(g)

        with pytest.raises(ValueError, match=r".*same number of static arguments.*"):
            arc.qpsol(f, g, x0=[0.0, 0.0], lba=0)

        # Test inconsistent decision variables type
        def f(x):
            return x[0] ** 2 + x[1] ** 2

        def g(x):
            return x[0] + x[1] - 10

        with pytest.raises(ValueError, match=r".*Only scalar and vector.*"):
            arc.qpsol(f, g, x0=np.zeros((2, 2)), lba=0)



"""Numerical optimization algorithms"""

from ._least_squares import least_squares
from ._lm import LMResult, LMStatus, lm_solve
from ._minimize import minimize, nlp_solver
from ._qpsol import qpsol
from ._root import implicit, root

__all__ = [
    "nlp_solver",
    "minimize",
    "implicit",
    "root",
    "qpsol",
    "least_squares",
    "lm_solve",
    "LMStatus",
    "LMResult",
]



from typing import Protocol

import numpy as np

from archimedes import struct, StructConfig, UnionConfig

__all__ = [
    "GravityModel",
    "ConstantGravity",
    "ConstantGravityConfig",
    "PointGravity",
    "PointGravityCartesianConfig",
    "PointGravityLatLonConfig",
    "GravityConfig",
]


class GravityModel(Protocol):
    def __call__(self, p_E: np.ndarray) -> np.ndarray:
        """Gravitational acceleration at the body CM in the earth frame E

        Args:
            p_E: position vector relative to the earth frame E [m]

        Returns:
            g_E: gravitational acceleration in earth frame E [m/s^2]
        """


@struct
class ConstantGravity:
    """Constant gravitational acceleration model

    This model assumes a constant gravitational acceleration vector
    in the +z direction (e.g. for a NED frame with "flat Earth" approximation)
    """

    g0: float = 9.81  # m/s^2

    def __call__(self, p_E: np.ndarray) -> np.ndarray:
        return np.hstack([0, 0, self.g0])


class ConstantGravityConfig(StructConfig, type="constant"):
    g0: float = 9.81  # m/s^2

    def build(self) -> ConstantGravity:
        return ConstantGravity(g0=self.g0)


def lla2eci(
    lat: float, lon: float, alt: float = 0.0, RE: float = 6.3781e6
) -> tuple[np.ndarray, np.ndarray]:
    """Convert latitude/longitude/altitude to Cartesian ECI coordinates.

    Args:
        lat: Latitude in degrees.
        lon: Longitude in degrees.
        alt: Altitude above the surface in meters.
        RE: Earth radius in meters.

    Returns:
        p_EN: Cartesian coordinates [m]
        R_EN: Quaternion matrix from NED to E
    """
    r = RE + alt  # Radius from Earth center [m]
    lat = np.deg2rad(lat)
    lon = np.deg2rad(lon)

    p_EN = r * np.array(
        [
            np.cos(lat) * np.cos(lon),
            np.cos(lat) * np.sin(lon),
            np.sin(lat),
        ]
    )

    # TODO: Use a built-in DCM function
    R_EN = np.array(
        [
            [-np.sin(lat) * np.cos(lon), -np.sin(lon), -np.cos(lat) * np.cos(lon)],
            [-np.sin(lat) * np.sin(lon), np.cos(lon), -np.cos(lat) * np.sin(lon)],
            [np.cos(lat), 0, -np.sin(lat)],
        ]
    )

    return p_EN, R_EN


@struct
class PointGravity:
    """Point mass gravitational acceleration model

    This model assumes a point mass at the origin of the inertial frame E
    """

    p_EN: np.ndarray  # Relative position of N with respect to E (measured in E) [m]
    R_EN: np.ndarray  # Quaternion from N to E
    mu: float = 3.986e14  # m^3/s^2

    def __call__(self, p_E: np.ndarray):
        r_E = self.p_EN + self.R_EN @ p_E
        r = np.linalg.norm(r_E)
        g_E = -self.mu * r_E / r**3
        return self.R_EN.T @ g_E


# Example of a base class config to demonstrate inheritance.
# Note that the "type" field is not specified here.
class PointGravityConfig(StructConfig):
    mu: float = 3.986e14  # m^3/s^2


class PointGravityCartesianConfig(PointGravityConfig, type="point_cartesian"):
    p_EN: np.ndarray  # Relative position of N with respect to E (measured in E) [m]
    R_EN: np.ndarray  # Quaternion from N to E

    def build(self) -> PointGravity:
        return PointGravity(self.p_EN, self.R_EN, mu=self.mu)


class PointGravityLatLonConfig(PointGravityConfig, type="point_latlon"):
    lat: float  # Latitude [deg]
    lon: float  # Longitude [deg]
    RE: float = 6.3781e6  # Earth radius [m]

    def build(self) -> PointGravity:
        p_EN, R_EN = lla2eci(self.lat, self.lon, RE=self.RE)
        return PointGravity(p_EN, R_EN, mu=self.mu)


GravityConfig = UnionConfig[
    ConstantGravityConfig,
    PointGravityCartesianConfig,
    PointGravityLatLonConfig,
]



"""Solving quadratic programming problems."""

from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Sequence, cast

import casadi as cs
import numpy as np

from archimedes import tree
from archimedes._core import (
    FunctionCache,
    SymbolicArray,
    _as_casadi_array,
    array,
    sym_like,
)

if TYPE_CHECKING:
    from ..typing import ArrayLike

__all__ = [
    "qpsol",
]


@tree.struct
class QPSolution:
    x: ArrayLike
    lam_a: ArrayLike


def qpsol(
    obj: Callable,
    constr: Callable,
    x0: ArrayLike,
    lba: ArrayLike | None = None,
    uba: ArrayLike | None = None,
    lam_a0: ArrayLike | None = None,
    args: Sequence = (),
    verbose: bool = False,
    name: str | None = None,
    warm_start: bool = True,
    **options,
) -> QPSolution:
    """Solve a quadratic programming problem

    This function solves a quadratic problem of the form:

    .. code-block:: text

        minimize        (1/2) x^T Q(p) x + c(p)^T x
        subject to      lba <= A(p)x <= uba

    where ``x`` represents decision variables and ``p`` represents parameters.
    The arrays ``Q``, ``c``, and ``A`` that define the convex quadratic program
    are determined by automatically differentiating the provided objective and
    constraint functions.  That is, if the objective and constraint functions do
    not define a convex quadratic program, the solver will solve the convex
    quadratic approximation to the provided problem, determined by linearization
    about the initial guess.

    Parameters
    ----------
    obj : callable
        Objective function to minimize with signature ``obj(x, *args)``.
        Must return a scalar value.
    constr : callable
        Constraint function with signature ``constr(x, *args)``.
        Must return a vector of constraint values where the constraints
        are interpreted as ``lba <= constr(x, *args) <= uba``.
    x0 : array-like
        Initial guess for the decision variables. This is used to determine
        the linearization point for the convex approximation and to warm-start
        the QP solver. If None, the initial guess will be set to zero.
    lba : array-like, optional
        Lower bounds for the constraints. If None, the lower bounds will be
        set to negative infinity.
    uba : array-like, optional
        Upper bounds for the constraints. If None, the upper bounds will be
        set to positive infinity.
    lam_a0 : array-like, optional
        Initial guess for the dual variables associated with the constraints.
        This is used to warm-start the QP solver.
    verbose : bool, default=False
        Print output from the solver, including number of iterations and convergence.
    name : str, optional
        Name for the resulting solver function. If None, a name will be
        generated based on the objective function name.
    warm_start : bool, default=True
        Whether to enable warm starting. Default is True.
    options : dict
        Additional options passed to the underlying QP solver (OSQP).

    Returns
    -------
    solution : QPSolution

        A named tuple containing the solution to the QP problem, including:
            - ``x``: The optimal decision variables.
            - ``lam_a``: The dual variables associated with the constraints.

    Notes
    -----
    When to use this function:
    - For solving convex quadratic optimization problems efficiently
    - For embedding QP solvers in larger computational graphs
    - As part of model predictive control (MPC) or trajectory optimization

    The solution to the quadratic program is unique, so the initial guess is less
    important than for more general nonlinear programming.  The exception is when
    the QP is the convex approximation to a nonlinear program, in which case the
    initial guess is used as the linearization point.

    This function supports code generation, but requires linking the OSQP C library
    to the generated code.

    Edge cases:
    - If the objective function is not convex (i.e., the Hessian is not positive
    semidefinite), OSQP may fail to find a solution.
    - For problems with equality constraints, set the same value for both the
    lower and upper bounds.
    - Currently only supports scalar and vector decision variables, not matrices.

    Examples
    --------
    >>> import numpy as np
    >>> import archimedes as arc
    >>>
    >>> # Define a simple QP: minimize x^2 + y^2 subject to x + y >= 1
    >>> def obj(z):
    ...     return np.dot(z, z)
    ...
    >>> def constr(z):
    ...     return z[0] + z[1]
    ...
    >>> # Create initial guess and constraint bounds
    >>> z0 = np.array([0.0, 0.0])
    >>> lba = 1.0  # Lower bound for x + y >= 1
    >>>
    >>> # Create and solve the QP
    >>> sol = arc.qpsol(obj, constr, z0, lba=lba)
    >>> print(f"Optimal solution: x = {sol.x}")
    >>> print(f"Dual variables: λ = {sol.lam_a}")

    See Also
    --------
    minimize : More general interface for nonlinear optimization problems
    root : Function for solving systems of nonlinear equations
    """

    options = {
        "warm_start_primal": warm_start,
        "warm_start_dual": warm_start,
        "osqp": {
            "verbose": verbose,
            **options,
        },
    }

    if not isinstance(obj, FunctionCache):
        obj = FunctionCache(obj)

    if not isinstance(constr, FunctionCache):
        constr = FunctionCache(constr)

    # Check that arguments and static arguments are the same for both functions
    if not len(obj.arg_names) == len(constr.arg_names):
        raise ValueError(
            "Objective and constraint functions must have the same number of arguments"
        )

    if not len(obj.static_argnums) == len(constr.static_argnums):
        raise ValueError(
            "Objective and constraint functions must have the same number of "
            "static arguments"
        )

    x0 = array(x0)
    ret_shape = x0.shape
    ret_dtype = x0.dtype

    # TODO: Shape checking for bounds
    if len(ret_shape) > 1 and ret_shape[1] > 1:
        raise ValueError(
            "Only scalar and vector decision variables are supported. "
            f"Got shape {ret_shape}"
        )

    # Flatten the arguments into a single array `p`.  If necessary,
    # this could be unflattened using `_param_struct.unravel`
    p_flat, unravel = tree.ravel(args)
    p_flat_sym = sym_like(p_flat, name="p", kind="MX")
    sym_args = unravel(p_flat_sym)

    # Define a state variable for the optimization
    x = sym_like(x0, name="x", kind="MX")
    f = obj(x, *sym_args)
    g = constr(x, *sym_args)

    # For type checking
    f = cast(SymbolicArray, f)
    g = cast(SymbolicArray, g)

    qp = {
        "x": x._sym,
        "f": f._sym,
        "g": g._sym,
    }

    if p_flat.size != 0:
        qp["p"] = p_flat_sym._sym

    solver = cs.qpsol("qp_solver", "osqp", qp, options)

    # Setup for evaluating the QP solver
    if lba is None:
        lba = -np.inf * np.ones(g.shape)

    if uba is None:
        uba = np.inf * np.ones(g.shape)

    # Before calling the CasADi solver interface, make sure everything is
    # either a CasADi symbol or a NumPy array
    x0, lba, uba = map(
        _as_casadi_array,
        (x0, lba, uba),
    )

    kwargs = {
        "x0": x0,
        "lbg": lba,
        "ubg": uba,
    }

    if p_flat.size != 0:
        kwargs["p"] = _as_casadi_array(p_flat)

    # Add dual variables if provided
    if lam_a0 is not None:
        kwargs["lam_g0"] = _as_casadi_array(lam_a0)

    # The return is a dict with keys `f`, `g`, `x` (and dual variables)
    sol = solver(**kwargs)

    # Return the solution and dual variables
    x = array(sol["x"], dtype=ret_dtype).reshape(ret_shape)  # type: ignore[assignment]
    lam_a = array(sol["lam_g"], dtype=ret_dtype).reshape(g.shape)

    return QPSolution(x=x, lam_a=lam_a)



# ruff: noqa: N802, N803, N806, E741

import numpy as np
import pytest

import archimedes as arc
from archimedes.sysid import Timeseries


class TestTimeseries:
    """Unit tests for Timeseries class."""

    def test_valid_construction(self):
        """Test successful construction with valid inputs."""
        # Time vector
        ts = np.array([0.0, 0.1, 0.2, 0.3])
        nt = len(ts)

        # Input signals (2 inputs)
        us = np.random.randn(2, nt)

        # Output measurements (3 outputs)
        ys = np.random.randn(3, nt)

        # Should construct successfully
        timeseries = Timeseries(ts=ts, us=us, ys=ys)

        assert np.array_equal(timeseries.ts, ts)
        assert np.array_equal(timeseries.us, us)
        assert np.array_equal(timeseries.ys, ys)

        # Test indexing
        timeseries_slice = timeseries[:2]
        assert timeseries_slice.ts.shape == (2,)
        assert timeseries_slice.us.shape == (2, 2)
        assert timeseries_slice.ys.shape == (3, 2)

        assert np.array_equal(timeseries_slice.ts, ts[:2])
        assert np.array_equal(timeseries_slice.us, us[:, :2])
        assert np.array_equal(timeseries_slice.ys, ys[:, :2])

    def test_minimal_dimensions(self):
        """Test construction with minimal valid dimensions."""
        # Single time point
        ts = np.array([0.0])

        # Single input, single output
        us = np.array([[1.0]])  # shape (1, 1)
        ys = np.array([[2.0]])  # shape (1, 1)

        timeseries = Timeseries(ts=ts, us=us, ys=ys)

        assert timeseries.ts.shape == (1,)
        assert timeseries.us.shape == (1, 1)
        assert timeseries.ys.shape == (1, 1)

    def test_multiple_dimensions(self):
        """Test construction with multiple inputs/outputs."""
        ts = np.linspace(0, 1, 100)
        nt = len(ts)

        # 5 inputs, 3 outputs
        us = np.random.randn(5, nt)
        ys = np.random.randn(3, nt)

        timeseries = Timeseries(ts=ts, us=us, ys=ys)

        assert timeseries.ts.shape == (100,)
        assert timeseries.us.shape == (5, 100)
        assert timeseries.ys.shape == (3, 100)

    def test_time_vector_not_1d_error(self):
        """Test error when time vector is not 1D."""
        # 2D time vector (invalid)
        ts = np.array([[0.0, 0.1], [0.2, 0.3]])
        us = np.random.randn(2, 4)
        ys = np.random.randn(1, 4)

        with pytest.raises(ValueError, match="Time vector must be one-dimensional"):
            Timeseries(ts=ts, us=us, ys=ys)

    def test_time_vector_0d_error(self):
        """Test error when time vector is 0D scalar."""
        ts = np.array(0.0)  # 0D array
        us = np.random.randn(1, 1)
        ys = np.random.randn(1, 1)

        with pytest.raises(ValueError, match="Time vector must be one-dimensional"):
            Timeseries(ts=ts, us=us, ys=ys)

    def test_time_vector_3d_error(self):
        """Test error when time vector is 3D."""
        ts = np.random.randn(2, 2, 2)  # 3D array
        us = np.random.randn(1, 8)
        ys = np.random.randn(1, 8)

        with pytest.raises(ValueError, match="Time vector must be one-dimensional"):
            Timeseries(ts=ts, us=us, ys=ys)

    def test_outputs_not_2d_error(self):
        """Test error when outputs are not 2D."""
        ts = np.array([0.0, 0.1, 0.2])
        us = np.random.randn(2, 3)

        # 1D outputs (invalid)
        ys = np.array([1.0, 2.0, 3.0])

        with pytest.raises(
            ValueError,
            match="Output measurements must be two-dimensional with shape \\(ny, nt\\)",
        ):
            Timeseries(ts=ts, us=us, ys=ys)

    def test_outputs_0d_error(self):
        """Test error when outputs are 0D."""
        ts = np.array([0.0])
        us = np.random.randn(1, 1)
        ys = np.array(5.0)  # 0D array

        with pytest.raises(
            ValueError,
            match="Output measurements must be two-dimensional with shape \\(ny, nt\\)",
        ):
            Timeseries(ts=ts, us=us, ys=ys)

    def test_outputs_3d_error(self):
        """Test error when outputs are 3D."""
        ts = np.array([0.0, 0.1])
        us = np.random.randn(1, 2)
        ys = np.random.randn(2, 2, 2)  # 3D array

        with pytest.raises(
            ValueError,
            match="Output measurements must be two-dimensional with shape \\(ny, nt\\)",
        ):
            Timeseries(ts=ts, us=us, ys=ys)

    def test_inputs_not_2d_error(self):
        """Test error when inputs are not 2D."""
        ts = np.array([0.0, 0.1, 0.2])
        ys = np.random.randn(1, 3)

        # 1D inputs (invalid)
        us = np.array([1.0, 2.0, 3.0])

        with pytest.raises(
            ValueError,
            match="Input signals must be two-dimensional with shape \\(nu, nt\\)",
        ):
            Timeseries(ts=ts, us=us, ys=ys)

    def test_inputs_0d_error(self):
        """Test error when inputs are 0D."""
        ts = np.array([0.0])
        ys = np.random.randn(1, 1)
        us = np.array(3.0)  # 0D array

        with pytest.raises(
            ValueError,
            match="Input signals must be two-dimensional with shape \\(nu, nt\\)",
        ):
            Timeseries(ts=ts, us=us, ys=ys)

    def test_inputs_3d_error(self):
        """Test error when inputs are 3D."""
        ts = np.array([0.0, 0.1])
        ys = np.random.randn(1, 2)
        us = np.random.randn(2, 2, 2)  # 3D array

        with pytest.raises(
            ValueError,
            match="Input signals must be two-dimensional with shape \\(nu, nt\\)",
        ):
            Timeseries(ts=ts, us=us, ys=ys)

    def test_time_outputs_size_mismatch_error(self):
        """Test error when time vector size doesn't match outputs time dimension."""
        ts = np.array([0.0, 0.1, 0.2])  # 3 time points
        us = np.random.randn(2, 3)  # 3 time points (correct)
        ys = np.random.randn(1, 4)  # 4 time points (incorrect)

        with pytest.raises(
            ValueError,
            match="Time vector size must match the number of time points in ys",
        ):
            Timeseries(ts=ts, us=us, ys=ys)

    def test_time_inputs_size_mismatch_error(self):
        """Test error when time vector size doesn't match inputs time dimension."""
        ts = np.array([0.0, 0.1, 0.2])  # 3 time points
        ys = np.random.randn(1, 3)  # 3 time points (correct)
        us = np.random.randn(2, 5)  # 5 time points (incorrect)

        with pytest.raises(
            ValueError,
            match="Time vector size must match the number of time points in us",
        ):
            Timeseries(ts=ts, us=us, ys=ys)

    def test_inputs_outputs_size_mismatch_error(self):
        """Test error when inputs and outputs have different time dimensions."""
        ts = np.array([0.0, 0.1, 0.2])  # 3 time points
        us = np.random.randn(2, 3)  # 3 time points
        ys = np.random.randn(1, 2)  # 2 time points (mismatch)

        # Should fail on outputs check first
        with pytest.raises(
            ValueError,
            match="Time vector size must match the number of time points in ys",
        ):
            Timeseries(ts=ts, us=us, ys=ys)

    def test_tree_functionality(self):
        """Test that Timeseries works with tree operations."""
        ts = np.array([0.0, 0.1, 0.2])
        us = np.random.randn(2, 3)
        ys = np.random.randn(1, 3)

        timeseries = Timeseries(ts=ts, us=us, ys=ys)

        # Test flattening
        flat, _ = arc.tree.flatten(timeseries)
        assert len(flat) == 3  # ts, us, ys

        # Test mapping
        scaled = arc.tree.map(lambda x: x * 2, timeseries)
        assert np.allclose(scaled.ts, ts * 2)
        assert np.allclose(scaled.us, us * 2)
        assert np.allclose(scaled.ys, ys * 2)

    def test_replace_method(self):
        """Test the replace method for creating modified copies."""
        ts = np.array([0.0, 0.1, 0.2])
        us = np.random.randn(2, 3)
        ys = np.random.randn(1, 3)

        timeseries = Timeseries(ts=ts, us=us, ys=ys)

        # Replace time vector
        new_ts = np.array([1.0, 1.1, 1.2])
        new_timeseries = timeseries.replace(ts=new_ts)

        assert np.array_equal(new_timeseries.ts, new_ts)
        assert np.array_equal(new_timeseries.us, us)  # unchanged
        assert np.array_equal(new_timeseries.ys, ys)  # unchanged

        # Original should be unchanged (frozen dataclass)
        assert np.array_equal(timeseries.ts, ts)

    def test_replace_with_validation(self):
        """Test that replace method still validates dimensions."""
        ts = np.array([0.0, 0.1, 0.2])
        us = np.random.randn(2, 3)
        ys = np.random.randn(1, 3)

        timeseries = Timeseries(ts=ts, us=us, ys=ys)

        # Replace with incompatible dimensions should fail
        bad_ts = np.array([1.0, 1.1])  # Wrong size

        with pytest.raises(ValueError, match="Time vector size must match"):
            timeseries.replace(ts=bad_ts)

    def test_immutability(self):
        """Test that Timeseries instances are immutable (frozen)."""
        ts = np.array([0.0, 0.1, 0.2])
        us = np.random.randn(2, 3)
        ys = np.random.randn(1, 3)

        timeseries = Timeseries(ts=ts, us=us, ys=ys)

        # Should not be able to modify fields directly
        with pytest.raises(AttributeError):
            timeseries.ts = np.array([1.0, 1.1, 1.2])

    def test_edge_case_empty_time(self):
        """Test behavior with empty time vectors."""
        ts = np.array([])
        us = np.random.randn(1, 0)  # 0 time points
        ys = np.random.randn(1, 0)  # 0 time points

        # Should construct successfully
        timeseries = Timeseries(ts=ts, us=us, ys=ys)
        assert timeseries.ts.size == 0
        assert timeseries.us.shape == (1, 0)
        assert timeseries.ys.shape == (1, 0)

    def test_realistic_system_id_data(self):
        """Test with realistic system identification data dimensions."""
        # Common scenario: 10 seconds at 100 Hz sampling
        dt = 0.01
        t_end = 10.0
        ts = np.arange(0, t_end, dt)
        nt = len(ts)

        # CartPole system: 1 input (force), 4 outputs (x, θ, ẋ, θ̇)
        nu, ny = 1, 4
        us = np.random.randn(nu, nt)
        ys = np.random.randn(ny, nt)

        timeseries = Timeseries(ts=ts, us=us, ys=ys)

        assert timeseries.ts.shape == (nt,)
        assert timeseries.us.shape == (nu, nt)
        assert timeseries.ys.shape == (ny, nt)
        assert len(ts) == 1000  # 10s / 0.01s



import numpy as np

from archimedes import struct

__all__ = ["iir_step", "IIRFilter"]


def iir_step(
    u: float, u_prev: np.ndarray, y_prev: np.ndarray, b: np.ndarray, a: np.ndarray
) -> tuple[np.ndarray, np.ndarray, float]:
    """Perform one step of IIR filtering

    Applies an IIR filter using the difference equation:

    .. code-block:: none

        y[n] = (b[0] * u[n] + b[1] * u[n-1] + ... + b[M] * u[n-M]
                - a[1] * y[n-1] - ... - a[N] * y[n-N]) / a[0]

    Parameters
    ----------
    u : float
        Current input sample.
    u_prev : ndarray
        Previous input samples, with the most recent sample at index 0.
    y_prev : ndarray
        Previous output samples, with the most recent sample at index 0.
    b : ndarray
        Coefficients for the numerator, in order of descending degree.
    a : ndarray
        Coefficients for the denominator, in order of descending degree.

    Returns
    -------
    u_prev : ndarray
        Updated input history.
    y_prev : ndarray
        Updated output history.
    y : float
        Current output sample.

    Examples
    --------
    A simple moving average filter with a window size of 3:
    >>> b = np.array([1/3, 1/3, 1/3])  # Numerator coefficients
    >>> a = np.array([1.0, 0.0])       # Denominator coefficients
    >>> u_prev = np.zeros(len(b))       # Initialize input history
    >>> y_prev = np.zeros(len(a) - 1)   # Initialize output history
    >>> inputs = [1, 2, 2, 2, 2]
    >>> outputs = []
    >>> for u in inputs:
    ...     u_prev, y_prev, y = iir_step(u, u_prev, y_prev, b, a)
    ...     outputs.append(y)
    >>> outputs
    [1.0, 1.3333, 1.6666, 2.0, 2.0]
    """

    # Update input history
    u_prev[1:] = u_prev[:-1]
    u_prev[0] = u

    # Apply the difference equation
    y = (np.dot(b, u_prev) - np.dot(a[1:], y_prev[: len(a) - 1])) / a[0]

    # Update output history
    y_prev[1:] = y_prev[:-1]
    y_prev[0] = y

    return u_prev, y_prev, y


@struct
class IIRFilter:
    """Infinite Impulse Response (IIR) filter.

    Implements a discrete-time IIR filter using the difference equation:

    .. code-block:: none

        y[n] = (b[0] * u[n] + b[1] * u[n-1] + ... + b[M] * u[n-M]
                - a[1] * y[n-1] - ... - a[N] * y[n-N]) / a[0]

    where:
    - u[n] is the current input sample
    - y[n] is the current output sample
    - b are the numerator coefficients
    - a are the denominator coefficients

    Parameters
    ----------
    b : ndarray
        Coefficients for the numerator, in order of descending degree.
    a : ndarray
        Coefficients for the denominator, in order of descending degree.
    """

    b: np.ndarray  # Coefficients for the numerator
    a: np.ndarray  # Coefficients for the denominator

    @struct
    class State:
        u_prev: np.ndarray
        y_prev: np.ndarray

    def __call__(self, state: State, u: float) -> tuple[State, float]:
        u_prev, y_prev, y = iir_step(u, state.u_prev, state.y_prev, self.b, self.a)
        state = IIRFilter.State(u_prev=u_prev, y_prev=y_prev)
        return state, y

    @property
    def x0(self) -> State:
        """Create a new state with zeroed input and output history."""
        u_prev = np.zeros(len(self.b))
        y_prev = np.zeros(len(self.a) - 1)
        return IIRFilter.State(u_prev=u_prev, y_prev=y_prev)



"""Simple utility functions for demonstrating code generation"""

import os
import re

from IPython.display import Markdown, display


def cleanup():
    for filename in (
        "main",
        "iir_filter.c",
        "iir_filter.h",
        "iir_filter_kernel.c",
        "iir_filter_kernel.h",
        "fib.c",
        "fib.h",
        "fib_kernel.c",
        "fib_kernel.h",
        "output.txt",
        "compound_filter.c",
        "compound_filter.h",
        "compound_filter_kernel.c",
        "compound_filter_kernel.h",
    ):
        if os.path.exists(filename):
            os.remove(filename)


def display_text(content, show=True, language="c"):
    md = Markdown(f"```{language}\n{content}\n```")

    if show:
        display(md)
    else:
        return md


def extract_c_function(filename, function_name, show=True):
    """
    Extract a specific function from a C file using regex

    Args:
        filename: Path to the C file
        function_name: Name of the function to extract

    Returns:
        The extracted function as a string, or None if not found
    """
    try:
        with open(filename, "r") as f:
            source = f.read()

        # Pattern to match function definition
        pattern = re.compile(
            r".*" + re.escape(function_name) + r"\s*\([^)]*\)\s*{", re.MULTILINE
        )

        match = pattern.search(source)
        if not match:
            print(f"Function '{function_name}' not found in '{filename}'")
            return None

        # Find the start position of the function
        start_pos = match.start()

        # Count braces to find the end of the function
        brace_count = 0
        in_function = False
        end_pos = start_pos

        for i in range(start_pos, len(source)):
            if source[i] == "{":
                in_function = True
                brace_count += 1
            elif source[i] == "}":
                brace_count -= 1
                if in_function and brace_count == 0:
                    end_pos = i + 1
                    break

        # Extract the function code
        extracted_code = source[start_pos:end_pos]

        # Display in a markdown cell
        return display_text(extracted_code, show=show, language="c")

    except FileNotFoundError:
        print(f"File '{filename}' not found")
        return None



from typing import OrderedDict

import numpy as np

import archimedes as arc
from archimedes._core import FunctionCache

__all__ = [
    "rbf",
    "relu",
    "elu",
    "sigmoid",
    "swish",
    "adam",
    "dense",
]


def rbf(x):
    return np.exp(-(x**2))


def relu(x):
    return np.maximum(x, 0)


def elu(x, alpha=1.0):
    return np.where(x > 0, x, alpha * (np.exp(x) - 1))


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def swish(x):
    return x * sigmoid(x)


def adam(
    obj,
    x0,
    iters,
    args=(),
    static_argnames=None,
    lr=0.001,
    betas=(0.9, 0.999),
    eps=1e-8,
    weight_decay=0,
    print_every=None,
    callback=None,
    callback_iters=None,
):
    if not isinstance(obj, FunctionCache):
        obj = FunctionCache(obj, static_argnames=static_argnames)

    grad = arc.grad(obj)
    x = x0.copy()
    m = np.zeros_like(x0)
    v = np.zeros_like(x0)

    if callback_iters is not None and callback is None:

        def callback(i, x, J):
            return J

    history = OrderedDict({})
    for i in range(iters):
        g = grad(x, *args)
        # x = x - lr * g  # Basic gradient descent
        if weight_decay != 0:
            g = g + weight_decay * x
        m = betas[0] * m + (1 - betas[0]) * g
        v = betas[1] * v + (1 - betas[1]) * g**2
        mhat = m / (1 - betas[0] ** (i + 1))
        vhat = v / (1 - betas[1] ** (i + 1))
        x = x - lr * mhat / (vhat**0.5 + eps)

        if print_every is not None and (i + 1) % print_every == 0:
            print(f"Iter {i + 1}/{iters}: {obj(x, *args)}")

        if callback_iters is not None and (i + 1) % callback_iters == 0:
            J = obj(x, *args)
            history[i + 1] = callback(i + 1, x, J)

    return x, history


def glorot_normal(rng, n_in, n_out):
    scale = np.sqrt(2 / (n_in + n_out))
    return scale * rng.normal(size=(n_out, n_in))


def zero_initialization(rng, n_in, n_out):
    return np.zeros((n_out, n_in))


INITIALIZATIONS = {
    "glorot_normal": glorot_normal,
    "zero": zero_initialization,
}


def dense(
    num_inputs,
    width,
    depth,
    num_outputs,
    activation=np.tanh,
    activate_final=False,
    rng=None,
    kind="MX",
    initialization="glorot_normal",
):
    if rng is None:
        rng = np.random.default_rng()

    if initialization not in INITIALIZATIONS:
        raise ValueError(f"Unknown initialization {initialization}")
    init_weights = INITIALIZATIONS[initialization]

    weights = [init_weights(rng, num_inputs, width)]
    biases = [np.zeros(width)]

    for _ in range(depth - 1):
        weights.append(init_weights(rng, width, width))
        biases.append(np.zeros(width))

    weights.append(init_weights(rng, width, num_outputs))
    biases.append(np.zeros(num_outputs))

    p = np.concatenate([W.ravel() for W in weights] + biases)

    if activation is None:

        def activation(x):
            return x

    def forward(x, p):
        W_ = []
        b_ = []
        idx = 0

        for w in weights:
            W_.append(np.reshape(p[idx : idx + w.size], w.shape))
            idx += w.size
        for b in biases:
            b_.append(p[idx : idx + b.size, None])
            idx += b.size

        # Allowable input shapes are (n_in,) or (n_in, n_samples) for vectorization.
        # Cannot call with scalar inputs, even if num_inputs==1
        # Support vectorization by calling with (n_in, n_samples) arrays
        if x.ndim == 0:
            raise ValueError("Cannot evaluate dense network with scalar input.")

        # Either (n_in,) or (n_in, n_samples,)
        if x.shape[0] != num_inputs:
            raise ValueError(
                f"Input shape {x.shape} incompatible with num_inputs={num_inputs}"
            )

        if x.ndim == 1:
            num_samples = 1
            flatten_output = True
        else:
            num_samples = x.shape[1]
            flatten_output = False

        x = np.reshape(x, (num_inputs, num_samples))

        # Forward pass
        for i, (w, b) in enumerate(zip(W_, b_)):
            x = np.dot(w, x) + b
            if i < depth or activate_final:
                x = activation(x)

        # x has shape (num_outputs, num_samples)
        # If the input was 1D, flatten the output
        if flatten_output:
            x = np.reshape(x, (num_outputs,))

        return x

    forward = FunctionCache(
        forward,
        arg_names=("x", "p"),
        kind=kind,
    )

    return p, forward



# ruff: noqa: N802, N803, N806, N815, N816

import numpy as np
from scipy.signal import chirp

import archimedes as arc

from models import (
    TransferFunction,
    HammersteinWiener,
    TanhNonlinearity,
    CubicNonlinearity,
)

np.random.seed(0)


def generate_linear_oscillator():
    omega_n = 2.0  # Natural frequency [rad/s]
    zeta = 0.1  # Damping ratio [-]

    # Problem dimensions
    nx = 2  # state dimension (x₁, x₂)
    nu = 1  # input dimension (u)
    ny = 1  # output dimension (y = x₁)

    # Time vector
    t0, tf = 0.0, 20.0
    dt = 0.05
    ts = np.arange(t0, tf, dt)

    us = np.ones((nu, len(ts)))  # Constant input (step)

    def ode_rhs(t, x):
        x1, x2 = x[0], x[1]
        u = np.interp(t, ts, us[0, :])  # Interpolate input at time t

        x1_t = x2
        x2_t = -(omega_n**2) * x1 - 2 * zeta * omega_n * x2 + omega_n**2 * u

        return np.hstack([x2, x2_t])

    # Step response
    xs = arc.odeint(
        ode_rhs,
        t_span=(t0, tf),
        x0=np.zeros(nx),
        t_eval=ts,
        rtol=1e-8,
        atol=1e-10,
    )

    # Add measurement noise
    noise_std = 0.05
    ys = xs[:1, :] + np.random.normal(0, noise_std, (ny, len(ts)))

    data = np.vstack([ts.reshape(1, -1), us, ys])
    np.savetxt(
        "data/oscillator.csv",
        data.T,
        delimiter="\t",
        header="time\t\tu\t\t\ty",
        comments="",
        fmt="%.6f",
    )

    truth = {
        "omega_n": omega_n,
        "zeta": zeta,
        "noise_std": noise_std,
        "x0": np.zeros(nx),
    }
    np.savez(
        "data/oscillator_truth.npz",
        **truth,
    )


def generate_duffing_oscillator():
    alpha = 1.0  # linear stiffness
    beta = 5.0  # nonlinear stiffness
    delta = 0.02  # damping

    # Forcing parameters
    gamma = 8.0  # forcing amplitude
    omega = 0.5  # forcing frequency

    # Problem dimensions
    nx = 2  # state dimension (x₁, x₂)
    nu = 1  # input dimension (u)
    ny = 1  # output dimension (y = x₁)

    def u(t):
        """Forcing function: γcos(ωt)"""
        return gamma * np.cos(omega * t)

    def ode_rhs(t, x):
        """Duffing oscillator: ẍ + δẋ + αx + βx³ = γcos(ωt)"""
        x1, x2 = x[0], x[1]
        x2_t = -delta * x2 - alpha * x1 - beta * x1**3 + u(t)
        return np.hstack([x2, x2_t])

    t0, tf = 0.0, 40.0  # Longer simulation to capture rich dynamics
    dt = 0.02  # Smaller timestep for nonlinear system
    ts = np.arange(t0, tf, dt)

    x0 = np.array([1.0, 0.5])

    # Generate reference data
    xs = arc.odeint(
        ode_rhs,
        t_span=(t0, tf),
        x0=x0,
        t_eval=ts,
    )

    # Add measurement noise
    noise_std = 0.01
    ys = xs[:1, :] + np.random.normal(0, noise_std, (ny, len(ts)))
    us = u(ts).reshape(1, -1)  # Reshape input to match dimensions

    data = np.vstack([ts.reshape(1, -1), us, ys])
    np.savetxt(
        "data/duffing.csv",
        data.T,
        delimiter="\t",
        header="time\t\tu\t\t\ty",
        comments="",
        fmt="%.6f",
    )

    truth = {
        "alpha": alpha,
        "beta": beta,
        "delta": delta,
        "gamma": gamma,
        "omega": omega,
        "noise_std": noise_std,
        "x0": x0,
    }
    np.savez(
        "data/duffing_truth.npz",
        **truth,
    )


def generate_hammerstein_wiener():
    # Transfer function with zero and mixed pole structure
    # G(s) = K * (τ_z*s + 1) / ((τ_slow*s + 1) * (s² + 2*ζ*ω_n*s + ω_n²))

    K = 20.0  # DC gain
    tau_z = 0.5  # Zero time constant (lead compensation character)
    tau_slow = 2.0  # Slow pole time constant
    omega_n = 6.0  # Natural frequency of oscillatory pair
    zeta = 0.2  # Light damping (ζ = 0.3 gives nice oscillation)

    # Resulting transfer function:
    # G(s) = 2 * (0.5*s + 1) / ((10*s + 1) * (s² + 1.2*s + 4))

    num = np.array([K * tau_z, K])
    den = np.array(
        [
            tau_slow,
            1 + 2 * tau_slow * zeta * omega_n,
            tau_slow * omega_n**2 + 2 * zeta * omega_n,
            omega_n**2,
        ]
    )

    # Problem dimensions
    nx = len(den) - 1  # Number of states (order of the system)
    nu = 1  # Single input
    ny = 1  # Single output

    hw_sys = HammersteinWiener(
        input=TanhNonlinearity(saturation=1.5),
        lin_sys=TransferFunction(num, den),
        output=CubicNonlinearity(coeff=0.05),
    )
    noise_std = 0.01

    def hw_ode(t, x, u, sys):
        return sys.dynamics(t, x, u)

    def obs(t, x, u, sys):
        return sys.observation(t, x, u)

    vmap_obs = arc.vmap(obs, in_axes=(0, 1, 1, None))

    def simulate_system(sys, ts, us, x0, noise_std=0.0):
        """Simulate the Hammerstein-Wiener system."""

        def ode_rhs(t, x, sys):
            u = np.interp(t, ts, us[0]).reshape((nu,))
            return hw_ode(t, x, u, sys)

        xs_true = arc.odeint(
            ode_rhs,
            t_span=(ts[0], ts[-1]),
            x0=x0,
            args=(sys,),
            t_eval=ts,
            rtol=1e-8,
            atol=1e-10,
        )

        ys_ideal = vmap_obs(ts, xs_true, us, sys).T
        ys = ys_ideal + np.random.normal(0, noise_std, ys_ideal.shape)

        return arc.sysid.Timeseries(
            ts=ts,
            us=us,
            ys=ys,
        )

    # 1. Ladder response for steady-state

    # Time vector
    t0, tf = 0.0, 100.0
    dt = 0.1
    ts = np.arange(t0, tf, dt)

    u_ss = np.linspace(-5, 5, 10, endpoint=True)
    us = np.zeros((nu, len(ts)))  # Zero input
    for i, u in enumerate(u_ss):
        us[:, ts > i * 10.0] = u

    ladder_data = simulate_system(hw_sys, ts, us, x0=np.zeros(nx), noise_std=noise_std)

    np.savetxt(
        "data/part3_ladder.csv",
        np.vstack([ladder_data.ts, ladder_data.us, ladder_data.ys]).T,
        delimiter="\t",
        header="time\t\tu\t\t\ty",
        comments="",
        fmt="%.6f",
    )

    # 2. Step response for transient dynamics
    # Time vector
    t0, tf = 0.0, 10.0
    dt = 0.02
    ts = np.arange(t0, tf, dt)

    u0 = 2.0
    us = u0 * np.ones((nu, len(ts)))  # Constant input
    us[:, ts > 10.0] *= -u0  # Step down after 10 seconds
    step_data = simulate_system(hw_sys, ts, us, x0=np.zeros(nx), noise_std=noise_std)

    np.savetxt(
        "data/part3_step.csv",
        np.vstack([step_data.ts, step_data.us, step_data.ys]).T,
        delimiter="\t",
        header="time\t\tu\t\t\ty",
        comments="",
        fmt="%.6f",
    )

    # 3. Chirp input for frequency response
    def dyn_res(x, u):
        return hw_sys.dynamics(0, x, u)

    u0 = 1.0
    x0_chirp = arc.root(dyn_res, x0=np.zeros(nx), args=(u0,))

    us = u0 + 0.2 * chirp(ts, f0=0.01, f1=3.0, t1=tf, method="quadratic")
    chirp_data = simulate_system(
        hw_sys, ts, us.reshape((nu, -1)), x0=x0_chirp, noise_std=noise_std
    )

    np.savetxt(
        "data/part3_chirp.csv",
        np.vstack([chirp_data.ts, chirp_data.us, chirp_data.ys]).T,
        delimiter="\t",
        header="time\t\tu\t\t\ty",
        comments="",
        fmt="%.6f",
    )


if __name__ == "__main__":
    generate_linear_oscillator()
    print("Data generated")

    generate_duffing_oscillator()
    print("Data generated")

    generate_hammerstein_wiener()
    print("Hammerstein-Wiener data generated")



import warnings

warnings.warn(
    "Importing from archimedes.experimental.spatial is deprecated "
    "and will be removed in version 1.0. "
    "Please import from archimedes.spatial instead.",
    DeprecationWarning,
    stacklevel=2,
)

# Re-export everything from the new location
from archimedes.spatial import *



from typing import Callable
import numpy as np

from archimedes import struct, field
from archimedes import jac


@struct
class ContinuousEKF:
    dyn: Callable = field(static=True)  # ẋ = dyn(t, x, u)
    obs: Callable = field(static=True)  # y = obs(t, x, u)
    Q: np.ndarray  # Process noise covariance
    R: np.ndarray  # Measurement noise covariance
    unravel_x: Callable = field(static=True, default=None)
    unravel_y: Callable = field(static=True, default=None)

    @struct
    class State:
        x: np.ndarray
        P: np.ndarray

    @property
    def nx(self):
        return self.Q.shape[0]

    @property
    def ny(self):
        return self.R.shape[0]

    @property
    def R_inv(self):
        return np.linalg.inv(self.R)

    def dynamics(self, t, state: State, y, u):
        x = state.x
        P = np.reshape(state.P, (self.nx, self.nx))
        F = jac(self.dyn, argnums=1)(t, x, u)
        H = jac(self.obs, argnums=1)(t, x, u)
        K = P @ H.T @ self.R_inv
        x_t = self.dyn(t, x, u) + K @ (y - self.obs(t, x, u))
        P_t = F @ P + P @ F.T - K @ H @ P + self.Q
        P_t = P_t.flatten()
        return self.State(x=x_t, P=P_t)



import click
import matplotlib.pyplot as plt
import numpy as np
import serial
import serial.tools.list_ports

dt = 1e-4


def _get_usb_device():
    ports = serial.tools.list_ports.comports()

    for port in ports:
        # Filter by USB vendor/product ID or description
        if port.vid is not None:  # Has USB vendor ID (indicates USB device)
            print(f"Found USB device: {port.device} - {port.description}")
            return port.device

    raise FileNotFoundError("No USB device found")


def _wait_for_start_message(ser):
    """Wait for and parse the START message."""
    while True:
        line = ser.readline().decode().strip()
        if line.startswith("START"):
            sample_count, pwm_count, sample_rate, loop_cycles = map(
                int, line.split(",")[1:]
            )
            print(f"Receiving {sample_count} samples at sample rate {sample_rate}")
            if loop_cycles > 0:
                # Divide by 180 MHz to get execution time in µs
                print(f"Control loop time: {loop_cycles / 180:.4f} µs")
            return sample_count, pwm_count, sample_rate


def _collect_raw_samples(ser, sample_count):
    """Collect raw data samples from serial connection."""
    data = []
    samples_received = 0

    while samples_received < sample_count:
        line = ser.readline().decode().strip()
        if line == "END":
            break

        try:
            values = [int(x) for x in line.split(",")]
            data.append(values)
            samples_received += 1

            if samples_received % 100 == 0:
                print(f"Received {samples_received}/{sample_count} samples")

        except ValueError:
            continue

    return data


def _receive_single_stream(ser):
    """Receive and parse a single data stream from serial connection."""

    # Parse START message
    start_info = _wait_for_start_message(ser)
    sample_count, pwm_count, sample_rate = start_info

    # Collect raw data
    raw_data = _collect_raw_samples(ser, sample_count)

    # Convert to numpy array and apply unit conversions
    return _process_raw_data(raw_data, sample_rate, pwm_count)


def _process_raw_data(raw_data, sample_rate, pwm_count):
    """Convert raw data to proper units and format."""
    data = np.array(raw_data, dtype=float)

    # Convert sample count to timesteps
    data[:, 0] = sample_rate * dt * data[:, 0]

    # PWM duty cycle to [0-1]
    data[:, 1] /= pwm_count

    # Unit conversions: mV -> V, mA -> A, mdeg -> deg
    data[:, 2:] *= 1e-3

    return data


@click.command()
@click.option(
    "--port", default=None, help="Serial port (e.g., /dev/tty.usbmodem141303)"
)
@click.option("--save", default=None, help="Filename to save data as CSV")
def main(port=None, save=None):
    if port is None:
        # Determine the USB port by looking for /dev/tty.usbmodem.*
        port = _get_usb_device()

    with serial.Serial(port, baudrate=115200, timeout=5) as ser:
        print("Waiting for experiment data...")
        data = _receive_single_stream(ser)

    # Save to CSV
    if save:
        filename = f"data/{save}"
        header = "t [s]\t\tu [-]\t\tv [V]\t\ti [A]\t\tpos [deg]"
        np.savetxt(
            filename, data, delimiter="\t", fmt="%.6f", header=header, comments=""
        )

    _fig, ax = plt.subplots(4, 1, figsize=(7, 6), sharex=True)
    ax[0].plot(data[:, 0], 100 * data[:, 1])
    ax[0].grid()
    ax[0].set_ylabel(r"PWM duty [%]")

    data[:, 2] = np.where(data[:, 2] > 1e6, -1, data[:, 2])  # Remove erroneous spikes

    ax[1].plot(data[:, 0], data[:, 2])
    ax[1].grid()
    ax[1].set_ylabel(r"Motor voltage [V]")
    # ax[0].set_ylim([0, 15])

    ax[2].plot(data[:, 0], data[:, 3])
    ax[2].grid()
    ax[2].set_ylabel(r"Motor current [A]")

    ax[3].plot(data[:, 0], data[:, 4])
    ax[3].grid()
    ax[3].set_ylabel(r"Position [deg]")

    ax[-1].set_xlabel("Time [s]")

    plt.show()


if __name__ == "__main__":
    main()



# ruff: noqa: N803, N806

from __future__ import annotations

import abc
from typing import Callable

import numpy as np
from scipy.linalg import cholesky  # For Cholesky decomposition

from archimedes import jac, tree

__all__ = [
    "KalmanFilterBase",
    "ExtendedKalmanFilter",
    "UnscentedKalmanFilter",
]


# # NOTE: Currently unused, but necessary for MLE calculations
# @compile(kind="SX")
# def _spd_logdet(A):
#     """slogdet for SPD matrices using Cholesky decomposition"""
#     L = np.linalg.cholesky(A)  # A = L @ L.T

#     # det(A) = det(L)^2 = (product of diagonal elements of L)^2
#     # log(det(A)) = 2 * sum(log(diagonal elements of L))
#     return 2.0 * sum(np.log(L[i, i]) for i in range(A.shape[0]))


def _default_missing(y):
    return False


@tree.struct
class KalmanFilterBase(metaclass=abc.ABCMeta):
    """Abstract base class for Kalman filter implementations.

    This class defines the common interface for Kalman filters used in system
    identification and state estimation.

    All Kalman filter implementations follow the discrete-time state-space model:

    .. code-block:: text

        x[k+1] = f(t[k], x[k], *args) + w[k]    (dynamics)
        y[k]   = h(t[k], x[k], *args) + v[k]    (observations)

    where ``w[k] ~ N(0, Q)`` is process noise and ``v[k] ~ N(0, R)`` is
    measurement noise.

    Parameters
    ----------
    dyn : callable
        Dynamics function with signature ``f(t, x, *args)`` that returns the
        predicted state at the next time step. Must be compatible with automatic
        differentiation for gradient-based filters.
    obs : callable
        Observation function with signature ``h(t, x, *args)`` that maps state
        to expected measurements. Must be compatible with automatic differentiation
        for gradient-based filters.
    Q : array_like
        Process noise covariance matrix of shape ``(nx, nx)`` where ``nx`` is
        the state dimension. Must be positive semi-definite.
    R : array_like
        Measurement noise covariance matrix of shape ``(ny, ny)`` where ``ny`` is
        the measurement dimension. Must be positive definite.
    missing : callable, optional
        Function to check if a value is missing (default: always returns False).
        The function should accept a single argument and return a boolean indicating
        whether the measurement is considered missing or corrupted.  The "correct"
        step will be skipped if the measurement is missing, and the filter step will
        return zero innovation with pure prediction for state and covariance.

    Attributes
    ----------
    nx : int
        State dimension, derived from the shape of ``Q``.
    ny : int
        Measurement dimension, derived from the shape of ``R``.

    Notes
    -----
    This class is decorated with ``@struct``, making it compatible
    with function transformations and enabling efficient automatic differentiation
    through the filter operations. The filter parameters can be modified using
    standard tree operations.

    Subclasses must implement the abstract ``step`` method that performs one
    complete filtering step (prediction + update). The implementation details
    depend on the specific filtering algorithm (EKF, UKF, etc.).

    The ``args`` parameter in function signatures allows passing additional
    arguments to both dynamics and observation functions, enabling time-varying
    parameters or external inputs.

    Examples
    --------
    >>> # Subclasses implement the filtering algorithm
    >>> class CustomFilter(KalmanFilterBase):
    ...     def step(self, t, x, y, P, args=None):
    ...         # Implementation-specific filtering logic
    ...         return x_new, P_new, innovation
    >>>
    >>> # Define system dynamics and observations
    >>> def f(t, x):
    ...     return np.array([x[0] + 0.1*x[1], 0.9*x[1]], like=x)
    >>>
    >>> def h(t, x):
    ...     return np.array([x[0]], like=x)
    >>>
    >>> # Create filter instance
    >>> Q = np.eye(2) * 0.01  # Process noise
    >>> R = np.array([[0.1]]) # Measurement noise
    >>> kf = CustomFilter(dyn=f, obs=h, Q=Q, R=R)
    >>>
    >>> # Filter properties
    >>> print(f"State dimension: {kf.nx}")
    >>> print(f"Measurement dimension: {kf.ny}")

    See Also
    --------
    ExtendedKalmanFilter : Extended Kalman Filter implementation
    UnscentedKalmanFilter : Unscented Kalman Filter implementation

    """

    dyn: Callable = tree.field(static=True)  # type: ignore[assignment]
    obs: Callable = tree.field(static=True)  # type: ignore[assignment]
    Q: np.ndarray
    R: np.ndarray

    missing: Callable[[np.ndarray], bool] = tree.field(
        default=_default_missing, static=True
    )  # type: ignore[assignment]

    @abc.abstractmethod
    def step(self, t, x, y, P, args=None):
        """Perform one step of the filter, combining prediction and update steps.

        This abstract method must be implemented by subclasses to define the
        specific filtering algorithm. It should perform both the prediction
        step (using the dynamics model) and the update step (incorporating
        the measurement).

        Parameters
        ----------
        t : float
            Current time step.
        x : array_like
            State vector of shape ``(nx,)``.
        y : array_like
            Measurement vector of shape ``(ny,)``.
        P : array_like
            State covariance matrix of shape ``(nx, nx)``.
        args : tuple, optional
            Additional arguments to pass to the dynamics and observation
            functions. Default is None.

        Returns
        -------
        x_new : ndarray
            Updated state estimate of shape ``(nx,)``.
        P_new : ndarray
            Updated state covariance matrix of shape ``(nx, nx)``.
        innovation : ndarray
            Measurement residual (innovation) of shape ``(ny,)``, computed
            as the difference between the actual measurement and the predicted
            measurement from the current state estimate.

        Notes
        -----
        The innovation sequence should have zero mean and known covariance
        (the innovation covariance) if the filter is performing optimally.
        This can be used for filter consistency checking and parameter tuning.
        """

    @property
    def nx(self):
        """State dimension derived from process noise covariance matrix."""
        return self.Q.shape[0]

    @property
    def ny(self):
        """Measurement dimension derived from measurement noise covariance matrix."""
        return self.R.shape[0]


@tree.struct
class ExtendedKalmanFilter(KalmanFilterBase):
    """Extended Kalman Filter for nonlinear state estimation.

    The Extended Kalman Filter (EKF) handles nonlinear dynamics and observation
    models by linearizing them around the current state estimate using Jacobian
    matrices. This makes it computationally efficient while providing reasonable
    performance for mildly nonlinear systems.

    The EKF uses first-order Taylor series approximations:

    .. code-block:: text

        F[k] ≈ ∂f/∂x |_{x[k|k-1]}    (dynamics Jacobian)
        H[k] ≈ ∂h/∂x |_{x[k|k-1]}    (observation Jacobian)

    These Jacobians are computed using automatic differentiation for accuracy and
    efficiency.

    Parameters
    ----------
    dyn : callable
        Dynamics function with signature ``f(t, x, *args)`` that returns the
        predicted state at the next time step. Must be compatible with automatic
        differentiation for gradient-based filters.
    obs : callable
        Observation function with signature ``h(t, x, *args)`` that maps state
        to expected measurements. Must be compatible with automatic differentiation
        for gradient-based filters.
    Q : array_like
        Process noise covariance matrix of shape ``(nx, nx)`` where ``nx`` is
        the state dimension. Must be positive semi-definite.
    R : array_like
        Measurement noise covariance matrix of shape ``(ny, ny)`` where ``ny`` is
        the measurement dimension. Must be positive definite.

    Notes
    -----
    The EKF is most suitable for systems where:

    - Nonlinearities are mild (locally approximately linear)
    - Computational efficiency is important
    - The system dynamics and observations are smooth and differentiable

    For highly nonlinear systems, consider using :class:`UnscentedKalmanFilter`
    which can better capture nonlinear transformations of uncertainty.

    The EKF algorithm consists of two steps:

    1. **Prediction**: Propagate state and covariance using linearized dynamics
    2. **Update**: Incorporate measurements using linearized observation model

    The linearization can introduce bias in the state estimates and may cause
    filter divergence if the nonlinearities are too strong or if the initial
    estimate is far from the true state.

    Examples
    --------
    >>> import numpy as np
    >>> from archimedes.observers import ExtendedKalmanFilter
    >>>
    >>> # Define nonlinear dynamics (simple pendulum)
    >>> def f(t, x):
    ...     dt = 0.1
    ...     return np.hstack([
    ...         x[0] + dt * x[1],                # position
    ...         x[1] - dt * 9.81 * np.sin(x[0])  # velocity
    ...     ])
    >>>
    >>> # Define observations
    >>> def h(t, x):
    ...     return x[0]
    >>>
    >>> # Create EKF
    >>> Q = np.eye(2) * 0.01
    >>> R = np.array([[0.1]])
    >>> ekf = ExtendedKalmanFilter(dyn=f, obs=h, Q=Q, R=R)
    >>>
    >>> # Filtering step
    >>> x = np.array([0.1, 0.0])  # Initial state
    >>> P = np.eye(2) * 0.1       # Initial covariance
    >>> y = np.array([0.12])      # Measurement
    >>> x_new, P_new, innov = ekf.step(0.0, x, y, P)

    See Also
    --------
    KalmanFilterBase : Abstract base class for all Kalman filters
    UnscentedKalmanFilter : Alternative for highly nonlinear systems

    References
    ----------
    .. [1] "Extended Kalman filter", Wikipedia.
           https://en.wikipedia.org/wiki/Extended_Kalman_filter
    .. [2] Gelb, A. "Applied Optimal Estimation." MIT Press, 1974.
    """

    def correct(self, t, x, y, P, args=None):
        """Perform the measurement update (correction) step of the EKF.

        This method applies the measurement update equations to incorporate
        new observations into the state estimate. It computes the observation
        Jacobian automatically and updates both the state estimate and its
        covariance matrix.

        Parameters
        ----------
        t : float
            Current time step.
        x : array_like
            Prior state estimate of shape ``(nx,)``, typically from the
            prediction step.
        y : array_like
            Measurement vector of shape ``(ny,)``.
        P : array_like
            Prior state covariance matrix of shape ``(nx, nx)``, typically
            from the prediction step.
        args : tuple, optional
            Additional arguments to pass to the observation function.
            Default is None.

        Returns
        -------
        x_post : ndarray
            Posterior (updated) state estimate of shape ``(nx,)``.
        P_post : ndarray
            Posterior (updated) state covariance matrix of shape ``(nx, nx)``.
        innovation : ndarray
            Measurement residual of shape ``(ny,)``, computed as the
            difference between the actual measurement and the predicted
            measurement from the prior state estimate.

        Notes
        -----
        This method can be used independently of the prediction step, allowing
        for custom prediction schemes or processing multiple measurements at
        the same time step.

        The method computes the observation Jacobian ``H = ∂h/∂x`` using
        automatic differentiation, then applies the standard Kalman update
        equations:

        .. code-block:: text

            innovation = y - h(x_prior)
            S = H @ P_prior @ H.T + R
            K = P_prior @ H.T @ S^(-1)
            x_post = x_prior + K @ innovation
            P_post = (I - K @ H) @ P_prior

        Examples
        --------
        >>> # Apply correction with custom prediction
        >>> x_pred = custom_prediction_step(x_prev)
        >>> P_pred = custom_covariance_prediction(P_prev)
        >>> x_new, P_new, innov = ekf.correct(t, x_pred, y, P_pred)
        """
        h = self.obs
        R = self.R

        if args is None:
            args = ()

        H = jac(h, argnums=1)(t, x, *args)
        H = np.atleast_2d(H)  # Ensure H is a 2D array

        e = y - h(t, x, *args)  # Innovation or measurement residual
        S = H @ P @ H.T + R  # Innovation covariance
        K = P @ H.T @ np.linalg.inv(S)  # Kalman gain

        # Skip correction if measurement is missing
        is_missing = self.missing(y)
        e = np.where(is_missing, np.zeros_like(e), e)
        K = np.where(is_missing, np.zeros_like(K), K)

        x = x + K @ e  # Updated state estimate
        P = (np.eye(len(x)) - K @ H) @ P  # Updated state covariance

        return x, P, e

    def step(self, t, x, y, P, args=None):
        """Perform one complete EKF step (prediction + update).

        This method implements the full Extended Kalman Filter algorithm,
        performing both the prediction step (using the dynamics model) and
        the update step (incorporating the measurement) in sequence.

        Parameters
        ----------
        t : float
            Current time step.
        x : array_like
            Previous state estimate of shape ``(nx,)``.
        y : array_like
            Current measurement vector of shape ``(ny,)``.
        P : array_like
            Previous state covariance matrix of shape ``(nx, nx)``.
        args : tuple, optional
            Additional arguments to pass to both dynamics and observation
            functions. Default is None.

        Returns
        -------
        x_new : ndarray
            Updated state estimate of shape ``(nx,)``.
        P_new : ndarray
            Updated state covariance matrix of shape ``(nx, nx)``.
        innovation : ndarray
            Measurement residual of shape ``(ny,)``.

        Notes
        -----
        The method performs the following sequence:

        1. **Prediction Step**:
           - Compute dynamics Jacobian ``F = ∂f/∂x``
           - Propagate state: ``x_pred = f(t, x)``
           - Propagate covariance: ``P_pred = F @ P @ F.T + Q``

        2. **Update Step**:
           - Apply measurement correction using :meth:`correct`

        The automatic differentiation ensures that Jacobians are computed
        accurately and efficiently, even for complex nonlinear functions.
        """
        f = self.dyn
        Q = self.Q

        if args is None:
            args = ()

        F = jac(f, argnums=1)(t, x, *args)

        # Predict step
        x = f(t, x, *args)
        P = F @ P @ F.T + Q

        # Update step
        x, P, e = self.correct(t, x, y, P, args)

        return x, P, e


def _julier_sigma_points(x_center, P_cov, kappa):
    """Generate Julier sigma points"""
    n = len(x_center)

    # Cholesky decomposition - scipy returns upper triangular
    U = cholesky((n + kappa) * P_cov)

    sigmas = np.zeros((2 * n + 1, n))
    sigmas[0] = x_center

    for k in range(n):
        # Note: U[k] accesses the k-th row of the upper triangular matrix
        sigmas[k + 1] = x_center + U[k]  # x + U[k]
        sigmas[n + k + 1] = x_center - U[k]  # x - U[k]

    return sigmas


def _julier_weights(n, kappa):
    """Compute Julier weights"""
    Wm = np.full(2 * n + 1, 0.5 / (n + kappa))
    Wm[0] = kappa / (n + kappa)
    Wc = Wm.copy()  # For Julier, Wm and Wc are the same
    return Wm, Wc


@tree.struct
class UnscentedKalmanFilter(KalmanFilterBase):
    """Unscented Kalman Filter for highly nonlinear state estimation.

    The Unscented Kalman Filter (UKF) uses the unscented transform to handle
    nonlinear dynamics and observation models. Instead of linearizing like the
    EKF, it propagates a carefully chosen set of sample points (sigma points)
    through the nonlinear transformations to capture the mean and covariance
    to higher-order accuracy.

    The UKF provides several advantages over the Extended Kalman Filter:

    - **Higher accuracy**: Captures mean and covariance to 2nd order for any
      nonlinearity (3rd order for Gaussian distributions)
    - **No Jacobian computation**: Uses sigma points instead of derivatives
    - **Better stability**: Less prone to divergence with strong nonlinearities
    - **Symmetric treatment**: Handles positive and negative perturbations equally

    The filter uses the Julier sigma point generation scheme with symmetric
    scaling, providing 2n+1 sigma points for an n-dimensional state.

    Parameters
    ----------
    dyn : callable
        Dynamics function with signature ``f(t, x, *args)`` that returns the
        predicted state at the next time step. Must be compatible with automatic
        differentiation for gradient-based filters.
    obs : callable
        Observation function with signature ``h(t, x, *args)`` that maps state
        to expected measurements. Must be compatible with automatic differentiation
        for gradient-based filters.
    Q : array_like
        Process noise covariance matrix of shape ``(nx, nx)`` where ``nx`` is
        the state dimension. Must be positive semi-definite.
    R : array_like
        Measurement noise covariance matrix of shape ``(ny, ny)`` where ``ny`` is
        the measurement dimension. Must be positive definite.
    kappa : float, default=0.0
        Sigma point scaling parameter. Controls the spread of sigma points
        around the mean. Typical values:

        - ``kappa = 0`` (default): Minimal spread, good for most applications
        - ``kappa = 3 - n``: Traditional choice for Gaussian distributions
        - ``kappa > 0``: Larger spread, may help with highly nonlinear systems

        where ``n`` is the state dimension.

    Notes
    -----
    The UKF is most suitable for systems where:

    - Nonlinearities are moderate to strong
    - Higher accuracy is needed than EKF can provide
    - The computational cost (≈ 3x EKF) is acceptable
    - Robustness to modeling errors is important

    For mildly nonlinear systems where computational efficiency is critical,
    :class:`ExtendedKalmanFilter` may be more appropriate.

    The UKF algorithm consists of:

    1. **Sigma point generation**: Create 2n+1 points capturing mean/covariance
    2. **Prediction**: Propagate points through dynamics, compute statistics
    3. **Update**: Propagate points through observation model, apply correction

    The implementation uses the Julier symmetric sigma point scheme with equal
    weights for covariance calculations, providing excellent numerical properties.

    Examples
    --------
    >>> import numpy as np
    >>> from archimedes.observers import UnscentedKalmanFilter
    >>>
    >>> # Define highly nonlinear dynamics (attitude dynamics)
    >>> def f(t, x):
    ...     dt = 0.1
    ...     q0, q1, q2, q3, wx, wy, wz = x
    ...     # Quaternion kinematics (highly nonlinear)
    ...     dq = 0.5 * dt * np.hstack([
    ...         -wx*q1 - wy*q2 - wz*q3,
    ...         wx*q0 + wz*q2 - wy*q3,
    ...         wy*q0 - wz*q1 + wx*q3,
    ...         wz*q0 + wy*q1 - wx*q2
    ...     ])
    ...     return np.hstack([x[:4] + dq, x[4:]])  # Constant angular velocity
    >>>
    >>> # Nonlinear observation (magnetic field measurement)
    >>> def h(t, x):
    ...     q0, q1, q2, q3 = x[:4]
    ...     # Rotate reference vector by quaternion (highly nonlinear)
    ...     R11 = q0**2 + q1**2 - q2**2 - q3**2
    ...     R21 = 2*(q1*q2 + q0*q3)
    ...     R31 = 2*(q1*q3 - q0*q2)
    ...     return np.hstack([R11, R21, R31])  # Simplified magnetic field
    >>>
    >>> # Create UKF with appropriate scaling
    >>> Q = np.eye(7) * 0.01  # Process noise
    >>> R = np.eye(3) * 0.1   # Measurement noise
    >>> ukf = UnscentedKalmanFilter(dyn=f, obs=h, Q=Q, R=R, kappa=0.0)
    >>>
    >>> # Filtering step
    >>> x = np.array([1, 0, 0, 0, 0.1, 0.1, 0.1])  # Initial quaternion + rates
    >>> P = np.eye(7) * 0.1                         # Initial covariance
    >>> y = np.array([0.95, 0.2, 0.15])            # Magnetic field measurement
    >>> x_new, P_new, innov = ukf.step(0.0, x, y, P)

    See Also
    --------
    KalmanFilterBase : Abstract base class for all Kalman filters
    ExtendedKalmanFilter : More efficient alternative for mildly nonlinear systems

    References
    ----------
    .. [1] Julier, S.J. and Uhlmann, J.K. "The scaled unscented transformation."
           American Control Converence, 2002.
    .. [2] Wan, E.A. and Van der Merwe, R. "The unscented Kalman filter for
           nonlinear estimation." Proc. IEEE Adaptive Systems, 2000.
    .. [3] Van der Merwe, R. "Sigma-Point Kalman Filters for Probabilistic
           Inference in Dynamic State-Space Models." PhD thesis, 2004.
    """

    kappa: float = 0.0

    def step(self, t, x, y, P, args=None):
        """Perform one complete UKF step using the unscented transform.

        This method implements the full Unscented Kalman Filter algorithm,
        performing both the prediction step (using sigma points propagated
        through the dynamics model) and the update step (incorporating the
        measurement through the unscented transform) in sequence.

        The UKF uses sigma points to capture the mean and covariance of the
        state distribution to higher-order accuracy without requiring Jacobian
        computations. This makes it particularly suitable for highly nonlinear
        systems where the EKF's linearization may be inadequate.

        Parameters
        ----------
        t : float
            Current time step.
        x : array_like
            Previous state estimate of shape ``(nx,)``.
        y : array_like
            Current measurement vector of shape ``(ny,)``.
        P : array_like
            Previous state covariance matrix of shape ``(nx, nx)``.
        args : tuple, optional
            Additional arguments to pass to both dynamics and observation
            functions. Default is None.

        Returns
        -------
        x_new : ndarray
            Updated state estimate of shape ``(nx,)``.
        P_new : ndarray
            Updated state covariance matrix of shape ``(nx, nx)``.
        innovation : ndarray
            Measurement residual of shape ``(ny,)``, computed as the
            difference between the actual measurement and the predicted
            measurement from the unscented transform.

        Notes
        -----
        The method performs the following sequence:

        1. **Sigma Point Generation**: Create 2n+1 sigma points using the
           Julier symmetric scheme to capture the current state distribution.

        2. **Prediction Step**:
           - Propagate sigma points through dynamics function
           - Compute predicted state mean using weighted sigma points
           - Compute predicted covariance using weighted outer products
           - Regenerate sigma points around predicted state (critical for accuracy)

        3. **Update Step**:
           - Propagate predicted sigma points through observation function
           - Compute predicted measurement mean using weighted sigma points
           - Compute innovation covariance and cross-covariance matrices
           - Apply Kalman update equations using computed statistics

        The regeneration of sigma points around the predicted state in step 2
        is crucial for maintaining the accuracy of the unscented transform,
        especially for highly nonlinear dynamics.

        The weights are computed using the Julier scheme where mean weights
        ``Wm`` and covariance weights ``Wc`` are identical, providing symmetric
        treatment of positive and negative perturbations.

        Examples
        --------
        >>> # Complete UKF filtering step for nonlinear system
        >>> x_k, P_k, innov = ukf.step(t_k, x_prev, y_k, P_prev)
        >>>
        >>> # With additional function arguments
        >>> extra_params = (param1, param2)
        >>> x_k, P_k, innov = ukf.step(t_k, x_prev, y_k, P_prev, args=extra_params)
        """
        f = self.dyn
        h = self.obs
        Q = self.Q
        R = self.R
        kappa = self.kappa

        if args is None:
            args = ()

        n = len(x)

        # Generate weights
        Wm, Wc = _julier_weights(n, kappa)

        # Generate initial sigma points
        sigmas = _julier_sigma_points(x, P, kappa)

        # PREDICT STEP - propagate sigma points through dynamics
        sigmas_f = np.zeros((2 * n + 1, n))
        for i in range(2 * n + 1):
            sigmas_f[i] = f(t, sigmas[i], *args)

        # Predicted mean (unscented transform)
        x_pred = np.zeros(n)
        for i in range(2 * n + 1):
            x_pred += Wm[i] * sigmas_f[i]

        # Predicted covariance
        P_pred = Q.copy()
        for i in range(2 * n + 1):
            diff = sigmas_f[i] - x_pred
            P_pred += Wc[i] * np.outer(diff, diff)

        # Regenerate sigma points around predicted state (critical for accuracy)
        sigmas_pred = _julier_sigma_points(x_pred, P_pred, kappa)

        # UPDATE STEP - propagate predicted sigma points through measurement function
        dim_z = len(y)
        sigmas_h = np.zeros((2 * n + 1, dim_z))
        for i in range(2 * n + 1):
            sigmas_h[i] = h(t, sigmas_pred[i], *args)

        # Predicted measurement mean
        y_pred = np.zeros(dim_z)
        for i in range(2 * n + 1):
            y_pred += Wm[i] * sigmas_h[i]

        # Innovation covariance
        S = R.copy()
        for i in range(2 * n + 1):
            diff_y = sigmas_h[i] - y_pred
            S += Wc[i] * np.outer(diff_y, diff_y)

        # Cross-covariance
        Pxz = np.zeros((n, dim_z))
        for i in range(2 * n + 1):
            diff_x = sigmas_pred[i] - x_pred
            diff_y = sigmas_h[i] - y_pred
            Pxz += Wc[i] * np.outer(diff_x, diff_y)

        e = y - y_pred  # Innovation
        K = Pxz @ np.linalg.inv(S)  # Kalman gain

        # Skip correction if measurement is missing
        is_missing = self.missing(y)
        e = np.where(is_missing, np.zeros_like(e), e)
        K = np.where(is_missing, np.zeros_like(K), K)

        x_new = x_pred + K @ e  # Updated state estimate
        P_new = P_pred - K @ S @ K.T  # Updated covariance

        return x_new, P_new, e



import numpy as np

from archimedes import struct
from archimedes.experimental.signal import tf2ss


class TransferFunctionBase:
    @property
    def ss(self):
        """Convert transfer function to state-space representation."""
        return tf2ss(self.num, self.den)

    def dynamics(self, t, x, u):
        A, B, _, _ = self.ss
        x = np.atleast_1d(x)
        u = np.atleast_1d(u)
        return A @ x + B @ u

    def observation(self, t, x, u):
        _, _, C, D = self.ss
        x = np.atleast_1d(x)
        u = np.atleast_1d(u)
        return C @ x + D @ u


@struct
class TransferFunction(TransferFunctionBase):
    num: np.ndarray
    den: np.ndarray


@struct
class FirstOrderLag(TransferFunctionBase):
    gain: float
    tau: float

    @property
    def num(self):
        """Numerator coefficients of the transfer function."""
        return np.atleast_1d(self.gain)

    @property
    def den(self):
        """Denominator coefficients of the transfer function."""
        return np.hstack([self.tau, 1.0])


class StaticNonlinearity:
    def __call__(self, x):
        """Apply nonlinearity to input."""
        raise NotImplementedError("StaticNonlinearity must implement __call__ method.")


@struct
class HammersteinWiener:
    input: StaticNonlinearity
    lin_sys: TransferFunction
    output: StaticNonlinearity

    def dynamics(self, t, x, u):
        """Hammerstein-Wiener system dynamics."""
        # Linear system dynamics with input nonlinearity
        return self.lin_sys.dynamics(t, x, self.input(u))

    def observation(self, t, x, u):
        """Hammerstein-Wiener system observation."""
        # Apply output nonlinearity to linear system output
        return self.output(self.lin_sys.observation(t, x, u))


@struct
class TanhNonlinearity(StaticNonlinearity):
    saturation: float

    def __call__(self, x):
        """Apply tanh nonlinearity with saturation."""
        return self.saturation * np.tanh(x / self.saturation)


@struct
class CubicNonlinearity(StaticNonlinearity):
    coeff: np.ndarray

    def __call__(self, x):
        """Apply cubic nonlinearity."""
        return x + self.coeff * x**3



# ruff: noqa: N802, N803, N806, N815, N816
from __future__ import annotations

from typing import Protocol

import numpy as np

from archimedes import StructConfig, UnionConfig, struct
from archimedes._core.utils import find_equal

__all__ = [
    "AtmosphereModel",
    "AtmosphereConfig",
    "LinearAtmosphere",
    "LinearAtmosphereConfig",
    "StandardAtmosphere1976",
    "StandardAtmosphere1976Config",
]


class AtmosphereModel(Protocol):
    def __call__(self, Vt: float, alt: float) -> tuple[float, float]:
        """Compute Mach number and dynamic pressure at given altitude and velocity.

        Args:
            Vt: true airspeed [ft/s]
            alt: altitude [ft]

        Returns:
            mach: Mach number [-]
            qbar: dynamic pressure [lbf/ft²]
        """


@struct
class LinearAtmosphere:
    """
    Linear temperature gradient atmosphere model using barometric formula.

    Density varies as ρ = ρ₀(T/T₀)^n where n = g/(R·L) - 1
    for a linear temperature profile T = T₀(1 - βz).
    """

    g0: float = 32.17  # Gravitational acceleration [ft/s²]
    R0: float = 2.377e-3  # Density scale [slug/ft^3]
    gamma: float = 1.4  # Adiabatic index for air [-]
    Rs: float = 1716.3  # Specific gas constant for air [ft·lbf/slug-R]
    dTdz: float = 0.703e-5  # Temperature gradient scale [1/ft]
    Tmin: float = 390.0  # Minimum temperature [R]
    Tmax: float = 519.0  # Maximum temperature [R]
    max_alt: float = 35000.0  # Maximum altitude [ft]

    def __call__(self, Vt, alt):
        L = self.Tmax * self.dTdz  # Temperature gradient [°R/ft]
        n = self.g0 / (self.Rs * L) - 1  # Density exponent [-]
        Tfac = 1 - self.dTdz * alt  # Temperature factor [-]
        T = np.where(alt >= self.max_alt, self.Tmin, self.Tmax * Tfac)
        rho = self.R0 * Tfac**n
        mach = Vt / np.sqrt(self.gamma * self.Rs * T)
        qbar = 0.5 * rho * Vt**2
        return mach, qbar


class LinearAtmosphereConfig(StructConfig, type="linear"):
    def build(self) -> LinearAtmosphere:
        return LinearAtmosphere()


# Altitude [ft]
M_TO_FT = 3.28084
K_TO_R = 9 / 5
PA_TO_PSF = 0.02088
h_USSA1976 = M_TO_FT * np.array([0, 11000, 20000, 32000, 47000, 51000, 71000, 84852])
# Temperature [R]
T_USSA1976 = K_TO_R * np.array(
    [288.15, 216.65, 216.65, 228.65, 270.65, 270.65, 214.65, 186.95]
)
# Pressure [psf]
p_USSA1976 = PA_TO_PSF * np.array(
    [101325, 22632.06, 5474.89, 868.02, 110.91, 66.94, 3.96, 0.3734]
)
# Temperature lapse rate [R/ft]
L_USSA1976 = (K_TO_R / M_TO_FT) * np.array([-0.0065, 0, 0.001, 0.0028, 0, 0.0028, 0, 0])


@struct
class StandardAtmosphere1976(AtmosphereModel):
    """U.S. Standard Atmosphere, 1976"""

    g0: float = 32.174  # Gravity constant ft/s^2
    Rs: float = 1716.3  # Specific gas constant for air [ft·lbf/slug-R]
    gamma: float = 1.4  # Adiabatic index for air [-]

    def calc_p(self, alt: float) -> float:
        alt = np.fmax(0, alt)
        h1 = find_equal(alt, h_USSA1976, h_USSA1976)
        T1 = find_equal(alt, h_USSA1976, T_USSA1976)
        p1 = find_equal(alt, h_USSA1976, p_USSA1976)
        L = find_equal(alt, h_USSA1976, L_USSA1976)

        return np.where(
            L == 0,
            p1 * np.exp(-self.g0 * (alt - h1) / (self.Rs * T1)),
            p1 * (T1 / (T1 + L * (alt - h1))) ** (self.g0 / (self.Rs * L)),
        )

    def calc_T(self, alt: float) -> float:
        return np.interp(alt, h_USSA1976, T_USSA1976)

    def __call__(self, Vt: float, alt: float) -> tuple[float, float]:
        """Compute Mach number and dynamic pressure at given altitude and velocity."""
        p = self.calc_p(alt)
        T = self.calc_T(alt)
        rho = p / (self.Rs * T)
        amach = Vt / np.sqrt(self.gamma * self.Rs * T)  # Adiabatic Mach number
        qbar = 0.5 * rho * Vt**2
        return amach, qbar


class StandardAtmosphere1976Config(StructConfig, type="ussa1976"):
    g0: float = 32.174  # Gravity constant ft/s^2
    Rs: float = 1716.3  # Specific gas constant for air [ft·lbf/slug-R]
    gamma: float = 1.4  # Adiabatic index for air [-]

    def build(self) -> StandardAtmosphere1976:
        return StandardAtmosphere1976(
            g0=self.g0,
            Rs=self.Rs,
            gamma=self.gamma,
        )


AtmosphereConfig = UnionConfig[
    LinearAtmosphereConfig,
    StandardAtmosphere1976Config,
]



"""Utilities for discretizing a continuous ODE function"""
# ruff: noqa: N802, N806

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Callable, Protocol, cast

import numpy as np

from archimedes import scan
from archimedes._core import FunctionCache
from archimedes.optimize import implicit

if TYPE_CHECKING:
    from archimedes.typing import Tree

    DynArgs = tuple[float, np.ndarray, np.ndarray, Tree]
    DynFunc = Callable[[*DynArgs], np.ndarray]


__all__ = ["discretize"]

# # NOTE: This implementation fails because alpha[i] is not allowed
# # when i is symbolic.  In theory this would be a better way to do it
# # because it has fewer "call sites" of the RHS function
# def _discretize_rk4_scan(f, h):
#     # RK4 Butcher tableau
#     alpha = np.array([0, 1 / 2, 1 / 2, 1])
#     beta = np.array([
#         [0, 0, 0, 0],
#         [1 / 2, 0, 0, 0],
#         [0, 1 / 2, 0, 0],
#         [0, 0, 1, 0],
#     ])
#     c_sol = np.array([1 / 6, 1 / 3, 1 / 3, 1 / 6])

#     # Take a single RK4 step
#     def step(t0, x0, p):
#         def body_fun(k, i):
#             ti = t0 + h * alpha[i]
#             yi = x0 + h * np.dot(beta[i, :], k)
#             k[i] = f(ti, yi, p)
#             return k, np.array([])

#         k = np.zeros((4, x0.size), dtype=x0.dtype)
#         k, _ = scan(body_fun, k, length=4)
#         return x0 + h * np.dot(c_sol, k)

#     return step


class RKSolver(Protocol):
    func: DynFunc
    h: float

    def __call__(self, carry: DynArgs, i: int) -> tuple[DynArgs, np.ndarray]:
        """Implement a single integration step in the scan loop."""


@dataclasses.dataclass
class EulerSolver:
    func: DynFunc
    h: float

    def __call__(self, carry: DynArgs, i: int) -> tuple[DynArgs, np.ndarray]:
        t0, x0, u, p = carry
        x1 = x0 + self.h * self.func(t0, x0, u, p)
        new_carry = (t0 + self.h, x1, u, p)
        return new_carry, np.array([])  # Dummy output for scan


@dataclasses.dataclass
class RK4Solver:
    func: DynFunc
    h: float

    def __call__(self, carry: DynArgs, i: int) -> tuple[DynArgs, np.ndarray]:
        t0, x0, u, p = carry

        k1 = self.func(t0, x0, u, p)
        k2 = self.func(t0 + self.h / 2, x0 + self.h * k1 / 2, u, p)
        k3 = self.func(t0 + self.h / 2, x0 + self.h * k2 / 2, u, p)
        k4 = self.func(t0 + self.h, x0 + self.h * k3, u, p)
        x1 = x0 + self.h * (k1 + 2 * k2 + 2 * k3 + k4) / 6

        new_carry = (t0 + self.h, x1, u, p)
        return new_carry, np.array([])  # Dummy output for scan


@dataclasses.dataclass
class Radau5Solver:
    func: DynFunc
    h: float
    newton_solver: dataclasses.InitVar[str] = "fast_newton"
    solve: Callable[[np.ndarray, *DynArgs], np.ndarray] = dataclasses.field(init=False)

    @property
    def a(self) -> np.ndarray:
        return np.array(
            [
                [
                    (88 - 7 * np.sqrt(6)) / 360,
                    (296 - 169 * np.sqrt(6)) / 1800,
                    (-2 + 3 * np.sqrt(6)) / 225,
                ],
                [
                    (296 + 169 * np.sqrt(6)) / 1800,
                    (88 + 7 * np.sqrt(6)) / 360,
                    (-2 - 3 * np.sqrt(6)) / 225,
                ],
                [(16 - np.sqrt(6)) / 36, (16 + np.sqrt(6)) / 36, 1 / 9],
            ]
        )

    @property
    def b(self) -> np.ndarray:
        return np.array([(16 - np.sqrt(6)) / 36, (16 + np.sqrt(6)) / 36, 1 / 9])

    @property
    def c(self) -> np.ndarray:
        return np.array([(4 - np.sqrt(6)) / 10, (4 + np.sqrt(6)) / 10, 1])

    def __post_init__(self, newton_solver: str):
        func = cast(FunctionCache, self.func)
        sym_kind: str = func._kind

        # Define the residual function used in the Newton solver
        def F(k, t, y, u, p):
            n = y.size
            k = np.reshape(k, (3, n))
            f = np.zeros_like(k)

            ts = t + self.h * self.c
            ys = y + self.h * self.a @ k

            # TODO: Use scan here?
            for i in range(3):
                f[i] = self.func(ts[i], ys[i], u, p)

            f, k = np.reshape(f, (3 * n,)), np.reshape(k, (3 * n,))
            return f - k

        F = FunctionCache(
            F, kind=sym_kind, arg_names=["k", "t", "y", "u", "p"], return_names=["r"]
        )
        self.solve = implicit(F, solver=newton_solver)

    def __call__(self, carry: DynArgs, i: int) -> tuple[DynArgs, np.ndarray]:
        t, x0, u, p = carry
        n = x0.size
        # Solve the nonlinear system using Newton's method
        k0 = np.hstack([x0, x0, x0])
        k = self.solve(k0, t, x0, u, p)
        k = np.reshape(k, (3, n))
        t1 = t + self.h
        x1 = x0 + self.h * np.dot(self.b, k)
        new_carry = (t1, x1, u, p)
        return new_carry, np.array([])


def discretize(func=None, dt=None, method="rk4", n_steps=1, name=None, **options):
    """Convert continuous-time dynamics to discrete-time using numerical integration.

    Transforms a continuous-time ordinary differential equation (ODE) function
    into a discrete-time difference equation.

    The function implements numerical integration schemes while maintaining automatic
    differentiation compatibility for gradient-based optimization, extended Kalman
    filtering, etc. The resulting discrete-time function preserves the same calling
    convention as the original continuous-time dynamics.

    Given a continuous-time system:

    .. code-block:: text

        dx/dt = ode(t, x, u, params)

    this function returns a discrete-time approximation:

    .. code-block:: text

        x[k+1] = f(t[k], x[k], u[k], params)

    where the discrete-time function ``f`` integrates the continuous dynamics over
    the time interval ``dt`` using the specified numerical method.

    Can also be used as a decorator to convert a continuous-time function
    into a discrete-time function:

    .. code-block:: python

        @discretize(dt=0.1, method="rk4")
        def dyn(t, x, u, params):
            # ... continuous-time dynamics implementation

    Parameters
    ----------
    func : callable or FunctionCache
        Continuous-time dynamics function with signature ``func(t, x, u, params)``:

        - ``t`` : Current time (scalar)
        - ``x`` : State vector of shape ``(nx,)``
        - ``u`` : Input vector of shape ``(nu,)``
        - ``params`` : Parameters (any tree-compatible structure)

        Must return the time derivative ``dx/dt`` as an array of shape ``(nx,)``.
        The function can be a regular Python function or a pre-compiled
        :class:`FunctionCache` object.
    dt : float
        Sampling time step for discretization. Represents the time interval
        between discrete samples ``t[k+1] - t[k]``. Smaller values provide
        better accuracy but increase computational cost.
    method : str, optional
        Numerical integration method. Default is ``"rk4"``. Available methods:

        - ``"rk4"``: Fourth-order Runge-Kutta method. Explicit, O(h⁴) accuracy.
          Excellent balance of accuracy and computational efficiency for most
          systems. Recommended for well-behaved, non-stiff dynamics.

        - ``"radau5"``: Fifth-order Radau IIA implicit method. Implicit,
          A-stable with excellent stability properties. Suitable for stiff
          systems and when high accuracy is required. Involves solving
          nonlinear equations at each step using Newton's method.

    n_steps : int, optional
        Number of integration sub-steps within each sampling interval ``dt``.
        Default is 1. When ``n_steps > 1``, the integration step size becomes
        ``h = dt / n_steps``, improving accuracy for rapid dynamics or large
        sampling intervals. Useful for maintaining accuracy when ``dt`` is
        constrained by measurement availability rather than dynamics.
    name : str, optional
        Name for the resulting discrete-time function. If None, automatically
        generated as ``"{method}_{func.name}"``. Used for debugging and
        performance profiling of compiled functions.
    **options
        Additional method-specific options:

        For ``method="radau5"``:
            - ``newton_solver`` : str, default ``"fast_newton"``
              Newton solver for implicit equations. Other options inclue ``"kinsol"``
              and ``"newton"``.

    Returns
    -------
    discrete_func : FunctionCache
        Discrete-time dynamics function with signature
        ``discrete_func(t, x, u, params)`` that returns the next state
        ``x[k+1]`` given current state ``x[k]``. The function is automatically
        compiled for efficient evaluation and supports automatic differentiation
        for gradient computation.

    Notes
    -----
    **Method Selection Guide:**

    **RK4 Method** (``method="rk4"``):
        - **Best for**: Non-stiff systems, general-purpose discretization
        - **Advantages**: Fast, explicit, well-tested, excellent accuracy/cost ratio
        - **Limitations**: Can become unstable for stiff systems or large time steps
        - **Typical use**: Mechanical systems, mild nonlinearities, most applications

    **Radau5 Method** (``method="radau5"``):
        - **Best for**: Stiff systems, high accuracy requirements, DAE systems
        - **Advantages**: A-stable, handles stiff dynamics, high-order accuracy
        - **Limitations**: Computational overhead from implicit solve
        - **Typical use**: Chemical kinetics, electrical circuits, multiscale dynamics

    **Accuracy Considerations:**

    The discretization error depends on both the method order and the time step:

    - **RK4**: Error ∼ O(dt⁴), typically accurate for ``dt < T/10`` where ``T``
      is the fastest system time constant
    - **Radau5**: Error ∼ O(dt⁵), maintains accuracy even for moderate ``dt``
    - **Sub-stepping**: Use ``n_steps > 1`` when measurement rate limits ``dt``
      but dynamics require smaller integration steps

    **Automatic Differentiation:**

    The discretized function preserves automatic differentiation through the
    integration process, enabling efficient gradient computation for:

    - Parameter estimation via :func:`pem`
    - Sensitivity analysis and uncertainty quantification
    - Gradient-based optimal control

    **Performance Optimization:**

    The returned function is automatically compiled using CasADi's just-in-time
    compilation for efficient repeated evaluation. For system identification
    applications involving hundreds of function evaluations, this provides
    significant performance benefits over pure Python implementations.

    Examples
    --------
    **Basic Discretization:**

    >>> import numpy as np
    >>> import archimedes as arc
    >>>
    >>> # Define continuous-time harmonic oscillator
    >>> def oscillator(t, x, u, params):
    ...     omega = params["omega"]  # Natural frequency
    ...     return np.hstack([
    ...         x[1],                    # dx1/dt = x2 (velocity)
    ...         -omega**2 * x[0] + u[0]  # dx2/dt = -ω²x1 + u (acceleration)
    ...     ])
    >>>
    >>> # Discretize with 50ms sampling
    >>> dt = 0.05
    >>> oscillator_discrete = arc.discretize(oscillator, dt, method="rk4")
    >>>
    >>> # Simulate one step
    >>> t0 = 0.0
    >>> x0 = np.array([1.0, 0.0])  # Initial position and velocity
    >>> u0 = np.array([0.0])       # No input
    >>> params = {"omega": 2.0}    # 2 rad/s natural frequency
    >>>
    >>> x1 = oscillator_discrete(t0, x0, u0, params)
    >>> print(f"Next state: {x1}")
    Next state: [0.995004 -0.199334]

    **Stiff System with Radau5:**

    >>> # Stiff Van der Pol oscillator
    >>> def van_der_pol(t, x, u, params):
    ...     mu = params["mu"]  # Large damping parameter
    ...     return np.hstack([
    ...         x[1],
    ...         mu * (1 - x[0]**2) * x[1] - x[0]
    ...     ])
    >>>
    >>> # Use implicit method for stiff dynamics
    >>> dt = 0.1  # Larger time step acceptable with Radau5
    >>> vdp_discrete = arc.discretize(
    ...     van_der_pol, dt,
    ...     method="radau5",
    ...     newton_solver="fast_newton"
    ... )
    >>>
    >>> params_stiff = {"mu": 10.0}  # Stiff parameter
    >>> x0 = np.array([2.0, 0.0])
    >>> x1 = vdp_discrete(0.0, x0, np.array([0.0]), params_stiff)

    See Also
    --------
    archimedes.sysid.pem : Parameter estimation using discrete-time models
    odeint : Continuous-time integration for comparison and validation

    References
    ----------
    .. [1] Hairer, E., Nørsett, S.P., and Wanner, G. "Solving Ordinary Differential
           Equations I: Nonstiff Problems." 2nd edition, Springer, 1993.
    .. [2] Hairer, E. and Wanner, G. "Solving Ordinary Differential Equations II:
           Stiff and Differential-Algebraic Problems." 2nd edition, Springer, 1996.
    """

    def _discretize_impl(f, dt_val):
        """Implementation of the discretization logic."""
        h = dt_val / n_steps

        if not isinstance(f, FunctionCache):
            f = FunctionCache(f)

        if name is None:
            func_name = f"{method}_{f.name}"
        else:
            func_name = name

        h = dt_val / n_steps
        scan_fun: RKSolver = {
            "euler": EulerSolver,
            "rk4": RK4Solver,
            "radau5": Radau5Solver,
        }[method](f, h, **options)

        def step(t0, x0, u, p):
            carry = (t0, x0, u, p)

            if n_steps == 1:
                # Slightly faster compilation if scan is not used
                carry, _ = scan_fun(carry, 0)
            else:
                carry, _ = scan(scan_fun, carry, length=n_steps)

            _, xf, _, _ = carry
            return xf

        return FunctionCache(step, name=func_name, arg_names=["t", "x", "u", "p"])

    # Check if we're in decorator mode (func is None) or direct mode (func provided)
    if func is None:
        # Decorator mode: @discretize(dt=0.1, method="rk4")
        if dt is None:
            raise ValueError("dt must be specified")

        def decorator(f):
            # Decorator function that applies discretization to the wrapped function
            return _discretize_impl(f, dt)

        return decorator
    else:
        # Direct mode: discretize(func, dt=0.1, method="rk4", ...)
        if dt is None:
            raise ValueError("dt must be specified")

        return _discretize_impl(func, dt)

    # h = dt / n_steps

    # if not isinstance(func, FunctionCache):
    #     func = FunctionCache(func)

    # if name is None:
    #     name = f"{method}_{func.name}"

    # h = dt / n_steps
    # scan_fun = {
    #     "rk4": _rk4,
    #     "radau5": _radau5,
    # }[method](func, h, **options)

    # def step(t0, x0, u, p):
    #     carry = (t0, x0, u, p)

    #     if n_steps == 1:
    #         # Slightly faster compilation if scan is not used
    #         carry, _ = scan_fun(carry, 0)
    #     else:
    #         carry, _ = scan(scan_fun, carry, length=n_steps)

    #     _, xf, _, _ = carry
    #     return xf

    # return FunctionCache(step, name=name, arg_names=["t", "x", "u", "p"])



import casadi as cs

import archimedes as arc
from archimedes._core._array_impl import SymbolicArray, _as_casadi_array


@arc.compile
def find_equal(x, xp, yp):
    # Return the first value of yp[j] such that xp[j] >= x.

    # Since this is a compiled function, we can assume that both are symbolic arrays
    xp_cs = _as_casadi_array(xp)
    x_cs = _as_casadi_array(x)

    # Add a dummy value or CasADi will only go to the second-to-last element
    inf_ = cs.MX.inf()  # low only supports MX
    grid = cs.vcat([xp_cs, inf_])
    i_cs = cs.low(grid, x_cs)
    y_cs = yp._sym[i_cs, :]

    # FIXME: This could be a more general function, but it would need much
    # more careful shape checking.
    return SymbolicArray(y_cs, shape=yp[0].shape, dtype=yp.dtype)



# ruff: noqa: N802, N803, N806

import numpy as np
import pytest
from scipy.spatial.transform import Rotation as ScipyRotation

import archimedes as arc
from archimedes.spatial import (
    EulerAngles,
    euler_kinematics,
    euler_to_dcm,
    euler_to_quaternion,
)

np.random.seed(0)


TEST_SEQ = ["XYZ", "xyz", "ZYX", "zyx", "XYX", "zxz"]


def random_euler(wrapper=False, seq="xyz"):
    """Generate a random rotation"""
    angles = np.random.randn(3)
    if wrapper:
        angles = EulerAngles(angles, seq=seq)
    return angles


def _test_kinematics(wrapper):
    rpy = np.array([0.1, 0.2, 0.3])

    # Given roll-pitch-yaw rates, compute the body-frame angular velocity
    # using the rotation matrices directly.
    pqr = np.array([0.4, 0.5, 0.6])  # Roll, pitch, yaw rates
    C_roll = EulerAngles(rpy[0], "x").as_matrix()  # C_φ
    C_pitch = EulerAngles(rpy[1], "y").as_matrix()  # C_θ
    # Successively transform each rate into the body frame
    w_B_ex = np.array([pqr[0], 0.0, 0.0]) + C_roll @ (
        np.array([0.0, pqr[1], 0.0]) + C_pitch @ np.array([0.0, 0.0, pqr[2]])
    )

    # Use the Euler kinematics function to duplicate the transformation
    Hinv = euler_kinematics(rpy, inverse=True)
    w_B = Hinv @ pqr

    np.testing.assert_allclose(w_B, w_B_ex)

    if wrapper:
        rpy = EulerAngles(rpy)
        result = rpy.kinematics(w_B).array

    else:
        # Test the forward transformation
        H = euler_kinematics(rpy)
        result = H @ w_B

    np.testing.assert_allclose(pqr, result)


class TestEulerLowLevel:
    def test_euler_to_quaternion(self):
        angles = random_euler()
        for seq in TEST_SEQ:
            q1 = euler_to_quaternion(angles, seq=seq)
            q2 = ScipyRotation.from_euler(seq, angles).as_quat(scalar_first=True)
            np.testing.assert_allclose(q1, q2)

    def test_euler_to_dcm(self):
        angles = random_euler()
        for seq in TEST_SEQ:
            print(seq)
            R1 = euler_to_dcm(angles, seq=seq)
            R2 = ScipyRotation.from_euler(seq, angles).as_matrix()

            # NOTE: SciPy's rotation uses an "active" convention (rotating vectors
            # in a single frame), whereas our DCMs are "passive" (rotating between
            # frames).  Thus the matrices are transposes of each other.
            np.testing.assert_allclose(R1, R2.T)

    def test_kinematics(self):
        _test_kinematics(wrapper=False)

    def test_error_handling(self):
        # Invalid axis sequence
        with pytest.raises(ValueError, match="Expected axes from `seq`"):
            euler_to_dcm([0.1, 0.2, 0.3], "abc")

        # Invalid euler shape
        with pytest.raises(ValueError, match="Expected axis specification to be"):
            euler_to_dcm([0.1, 0.2, 0.3], "xyzyz")

        # Angles shape doesn't match sequence
        with pytest.raises(ValueError, match="For xyz sequence with 3 axes, `angles`"):
            euler_to_dcm([0.1, 0.2], "xyz")

        # Repeated axis in sequence
        with pytest.raises(ValueError, match="Expected consecutive axes"):
            euler_to_dcm([0.1, 0.2, 0.3], "xxz")


class TestEulerWrapper:
    def test_ops(self):
        angles = random_euler(wrapper=True)
        assert len(angles) == 3

        # __getitem__ and __len__
        short_angles = EulerAngles(angles[:2], seq=angles.seq[:2])
        assert len(short_angles) == 2

        # __iter__
        r, p, y = angles
        assert np.allclose(np.array([r, p, y]), angles.array)

    def test_identity(self):
        angles = EulerAngles.identity(seq="xyz")
        R = angles.as_matrix()
        np.testing.assert_allclose(R, np.eye(3))

    def test_rotate(self):
        seq = "xyz"
        angles = random_euler(wrapper=True, seq=seq)

        v = np.array([0.1, 0.2, 0.3])
        R = angles.as_matrix()
        w = R @ v

        R_scipy = ScipyRotation.from_euler(seq, angles.array)
        w_scipy = R_scipy.apply(v, inverse=True)

        assert np.allclose(w, w_scipy)

    def test_from_euler(self):
        angles = random_euler(wrapper=True, seq="xyz")

        # Test getting the same rotation back
        np.testing.assert_allclose(
            EulerAngles.from_euler(angles, seq=angles.seq).array, angles.array
        )

        R_scipy = ScipyRotation.from_euler(angles.seq, angles.array)

        for seq in TEST_SEQ:
            angles_out1 = EulerAngles.from_euler(angles, seq=seq).array
            angles_out2 = R_scipy.as_euler(seq=seq)
            np.testing.assert_allclose(angles_out1, angles_out2)

    def test_as_euler(self):
        angles = random_euler(wrapper=True, seq="xyz")

        # Test getting the same rotation back
        assert angles.as_euler(seq=angles.seq) is angles

        R_scipy = ScipyRotation.from_euler(angles.seq, angles.array)

        for seq in TEST_SEQ:
            angles_out1 = angles.as_euler(seq=seq).array
            angles_out2 = R_scipy.as_euler(seq=seq)
            np.testing.assert_allclose(angles_out1, angles_out2)

    def test_as_matrix(self):
        for seq in TEST_SEQ:
            angles = random_euler(wrapper=True, seq=seq)
            R1 = angles.as_matrix()
            R2 = ScipyRotation.from_euler(seq, angles.array).as_matrix()
            # NOTE: SciPy's rotation uses an "active" convention (rotating vectors
            # in a single frame), whereas our DCMs are "passive" (rotating between
            # frames).  Thus the matrices are transposes of each other.
            np.testing.assert_allclose(R1, R2.T)

    def test_inv(self):
        seq = "xyz"
        angles = random_euler(wrapper=True, seq=seq)

        angles_inv = angles.inv()
        R1 = angles.as_matrix()
        R2 = angles_inv.as_matrix()

        np.testing.assert_allclose(R1.T, R2)

    def test_kinematics(self):
        _test_kinematics(wrapper=True)

    def test_tree_ops(self):
        euler = EulerAngles([0.1, 0.2, 0.3], "xyz")
        flat, unflatten = arc.tree.ravel(euler)
        euler_restored = unflatten(flat)

        R1 = euler.as_matrix()
        R2 = euler_restored.as_matrix()
        assert np.allclose(R1, R2)



# ruff: noqa: N802, N803, N806

import numpy as np

import archimedes as arc
from archimedes.experimental.astro.orbital_elements import (
    KeplerElements,
    cartesian_to_kepler,
    kepler_to_cartesian,
)

# NOTE: All test data generated using MATLAB `kepler2ijk` and similar functions


# Compile this to check that symbolic/numeric evaluation works
@arc.compile
def round_trip(oe):
    cartesian = kepler_to_cartesian(oe)
    oe2 = cartesian_to_kepler(cartesian)
    return cartesian, oe2


# Helper function for testing Kepler -> Cartesian and back
def _test_round_trip(oe, r_ex, v_ex, rtol=1e-3):
    """
    Test the round trip conversion from Kepler elements to Cartesian and back.

    Args:
        oe: KeplerElements instance
        r_ex: Expected position vector
        v_ex: Expected velocity vector
        rtol: Relative tolerance for comparisons
    """
    # Convert Kepler elements to Cartesian and back
    cartesian, oe2 = round_trip(oe)
    r, v = cartesian.r, cartesian.v

    # Check if the Cartesian state matches the expected values
    assert np.allclose(r, r_ex, rtol=rtol)
    assert np.allclose(v, v_ex, rtol=rtol)

    # Check if the original elements match the converted ones
    assert np.isclose(oe.a, oe2.a, rtol=rtol)
    assert np.isclose(oe.e, oe2.e, rtol=rtol)
    assert np.isclose(oe.i, oe2.i, rtol=rtol)
    assert np.isclose(oe.omega, oe2.omega, rtol=rtol)
    assert np.isclose(oe.RAAN, oe2.RAAN, rtol=rtol)
    assert np.isclose(oe.nu, oe2.nu, rtol=rtol)


class TestKeplerElements:
    """Test suite for Keplerian elements conversion."""

    def test_circular_orbit(self):
        """Test conversion for a circular orbit."""
        # Create Kepler elements (a, e, i, omega, RAAN, nu)
        oe = KeplerElements(7000e3, 0.0, 0.0, 0.0, 0.0, np.deg2rad(45.0))

        # Expected position and velocity vectors
        r_ex = np.array([4949.7e3, 4949.7e3, 0.0])
        v_ex = np.array([-5.3359e3, 5.3359e3, 0.0])

        _test_round_trip(oe, r_ex, v_ex)

    def test_elliptical_orbit(self):
        """Test conversion for an elliptical orbit."""
        # KeplerElements(a, e, i, omega, RAAN, nu)
        oe = KeplerElements(
            10.0e6,
            0.5,
            np.deg2rad(30.0),
            np.deg2rad(90.0),
            np.deg2rad(60.0),
            np.deg2rad(180.0),
        )

        r_ex = np.array([11250.0e3, -6495.0e3, -7500.0e3])
        v_ex = np.array([1.8225e3, 3.1567e3, 0.0])

        _test_round_trip(oe, r_ex, v_ex)

        # Check conditional offset when Ω > 180 degrees
        oe = KeplerElements(
            10000.0e3,
            0.5,
            np.deg2rad(30.0),
            np.deg2rad(90.0),
            np.deg2rad(210.0),
            np.deg2rad(180.0),
        )

        r_ex = np.array([-6495.0e3, 11250.0e3, -7500.0e3])
        v_ex = np.array([-3.1567e3, -1.8225e3, 0.0])

        _test_round_trip(oe, r_ex, v_ex)

        # Check conditional offset when ω > 180 degrees
        oe = KeplerElements(
            10000.0e3,
            0.5,
            np.deg2rad(30.0),
            np.deg2rad(210.0),
            np.deg2rad(60.0),
            np.deg2rad(180.0),
        )

        r_ex = np.array([870.0e3, 14498.0e3, 3750.0e3])
        v_ex = np.array([-3.2788e3, -0.2115e3, 1.5784e3])

        _test_round_trip(oe, r_ex, v_ex)

        # Check conditional offset when ν > 180 degrees
        oe = KeplerElements(
            10000.0e3,
            0.5,
            np.deg2rad(30.0),
            np.deg2rad(210.0),
            np.deg2rad(60.0),
            np.deg2rad(200.0),
        )

        r_ex = np.array([-3581.0e3, 12568.0e3, 5419.0e3])
        v_ex = np.array([-3.0280e3, -2.5958e3, 0.7646e3])

        _test_round_trip(oe, r_ex, v_ex)

    def test_circular_polar_orbit(self):
        """Test conversion for a circular polar orbit."""
        oe = KeplerElements(8e6, 0.0, np.pi / 2, 0.0, 0.0, np.deg2rad(270.0))

        r_ex = np.array([0.0, 0.0, -8e6])
        v_ex = np.array([7.0587e3, 0.0, 0.0])

        _test_round_trip(oe, r_ex, v_ex)

        # Test conditional offset when arglat < 180 degrees
        oe = KeplerElements(8e6, 0.0, np.pi / 2, 0.0, 0.0, np.deg2rad(170.0))

        r_ex = np.array([-7.8785e6, 0.0, 1.3892e6])
        v_ex = np.array([-1.2257e3, 0.0, -6.9514e3])

        _test_round_trip(oe, r_ex, v_ex)

    def test_elliptical_equatorial_orbit(self):
        """Test conversion for an elliptical equatorial orbit."""
        oe = KeplerElements(12e6, 0.7, 0.0, np.deg2rad(45.0), 0.0, np.deg2rad(135.0))

        r_ex = np.array([-12.118e6, 0.0, 0.0])
        v_ex = np.array([-3.9946e3, -4.0757e3, 0.0])

        _test_round_trip(oe, r_ex, v_ex)

    def test_parabolic_orbit(self):
        """Test conversion for a parabolic orbit."""
        oe = KeplerElements(
            7e6,
            1.0,
            np.deg2rad(45.0),
            np.deg2rad(60.0),
            np.deg2rad(120.0),
            np.deg2rad(90.0),
        )

        r_ex = np.array([1.776e6, -12.975e6, 4.950e6])
        v_ex = np.array([4.8405e3, -5.6219e3, -1.3810e3])

        _test_round_trip(oe, r_ex, v_ex)

    def test_hyperbolic_orbit(self):
        """Test conversion for a hyperbolic orbit."""
        oe = KeplerElements(
            -15e6,
            1.5,
            np.deg2rad(60.0),
            np.deg2rad(30.0),
            np.deg2rad(300.0),
            np.deg2rad(120.0),
        )

        r_ex = np.array([-16.238e6, 65.625e6, 32.476e6])
        v_ex = np.array([-2.0172e3, 5.4904e3, 1.7290e3])

        _test_round_trip(oe, r_ex, v_ex)



"""Low-level functions for quaternion operations.

These functions are for conversions and kinematics and operate directly on
arrays rather than higher-level wrapper classes.
"""

# ruff: noqa: N806, N803, N815
from __future__ import annotations

from typing import cast

import numpy as np

from archimedes import array

from ._euler import _check_angles, _check_seq

__all__ = [
    "euler_to_quaternion",
    "quaternion_inverse",
    "quaternion_kinematics",
    "quaternion_multiply",
    "quaternion_to_dcm",
    "quaternion_to_euler",
]


def quaternion_inverse(q):
    """
    Inverse of a quaternion q = [w, x, y, z]
    """
    return np.array([q[0], -q[1], -q[2], -q[3]], like=q)


def quaternion_multiply(q1: np.ndarray, q2: np.ndarray) -> np.ndarray:
    """Multiply (compose) two quaternions

    Parameters
    ----------
    q1 : array_like, shape (4,)
        First quaternion [w1, x1, y1, z1]
    q2 : array_like, shape (4,)
        Second quaternion [w2, x2, y2, z2]

    Returns
    -------
    np.ndarray, shape (4,)
        Resulting quaternion from multiplication q = q1 * q2

    Notes
    -----
    This function uses the scalar-first convention for quaternions, i.e. a quaternion
    is represented as [w, x, y, z], where w is the scalar part.
    """
    w1, x1, y1, z1 = q1
    w2, x2, y2, z2 = q2

    return np.array(
        [
            w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2,
            w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2,
            w1 * y2 - x1 * z2 + y1 * w2 + z1 * x2,
            w1 * z2 + x1 * y2 - y1 * x2 + z1 * w2,
        ],
        like=q1,
    )


def _elementary_basis_index(axis: str) -> int:
    return {"x": 1, "y": 2, "z": 3}[axis.lower()]


# See https://github.com/scipy/scipy/blob/3ead2b543df7c7c78619e20f0cb6139e344a8866/scipy/spatial/transform/_rotation_cy.pyx#L358-L372  # ruff: noqa: E501
def _make_elementary_quat(axis: str, angle: float) -> np.ndarray:
    """Create a quaternion representing a rotation about a principal axis."""

    quat = np.hstack([np.cos(angle / 2), np.zeros(3)])
    axis_idx = _elementary_basis_index(axis)
    quat[axis_idx] = np.sin(angle / 2)

    return quat


# See https://github.com/scipy/scipy/blob/3ead2b543df7c7c78619e20f0cb6139e344a8866/scipy/spatial/transform/_rotation_cy.pyx#L376-L391  # ruff: noqa: E501
def _elementary_quat_compose(
    seq: str, angles: np.ndarray, intrinsic: bool
) -> np.ndarray:
    """Create a quaternion from a sequence of elementary rotations."""
    q = _make_elementary_quat(seq[0], angles[0])

    for idx in range(1, len(seq)):
        qi = _make_elementary_quat(seq[idx], angles[idx])
        if intrinsic:
            q = quaternion_multiply(q, qi)
        else:
            q = quaternion_multiply(qi, q)

    return q


def dcm_to_quaternion(matrix: np.ndarray) -> np.ndarray:
    """Create a Quaternion from a rotation matrix.

    If the rotation matrix is R_BA that represents the attitude of a body B
    relative to a frame A, then the resulting quaternion represents the attitude
    of the body B relative to frame A.

    Note that for the sake of symbolic computation, this method assumes that
    the input is a valid rotation matrix (orthogonal and determinant +1).

    Implementation based on SciPy and reference [1]_.

    Parameters
    ----------
    matrix : array_like, shape (3, 3)
        Quaternion matrix.

    Returns
    -------
    np.ndarray, shape (4,)
        Unit quaternion [w, x, y, z] corresponding to the rotation matrix,
        where w is the scalar part.

    References
    ----------
    .. [1] F. Landis Markley, "Unit Quaternion from Rotation Matrix",
            Journal of guidance, control, and dynamics vol. 31.2, pp.
            440-442, 2008.
    """
    matrix = cast(np.ndarray, array(matrix))
    if matrix.shape != (3, 3):
        raise ValueError("Rotation matrix must be 3x3")

    # Transpose of SciPy convention  (we assume passive rotations for coordinate
    # system transformations rather than active rotations of vectors)
    matrix = matrix.T

    t = np.linalg.trace(matrix)

    # If matrix[0, 0] is the largest diagonal element
    q0 = np.hstack(
        [
            1 - t + 2 * matrix[0, 0],
            matrix[0, 1] + matrix[1, 0],
            matrix[0, 2] + matrix[2, 0],
            matrix[2, 1] - matrix[1, 2],
        ]
    )

    # If matrix[1, 1] is the largest diagonal element
    q1 = np.hstack(
        [
            1 - t + 2 * matrix[1, 1],
            matrix[2, 1] + matrix[1, 2],
            matrix[0, 1] + matrix[1, 0],
            matrix[0, 2] - matrix[2, 0],
        ]
    )

    # If matrix[2, 2] is the largest diagonal element
    q2 = np.hstack(
        [
            1 - t + 2 * matrix[2, 2],
            matrix[0, 2] + matrix[2, 0],
            matrix[2, 1] + matrix[1, 2],
            matrix[1, 0] - matrix[0, 1],
        ]
    )

    # If t is the largest diagonal element
    q3 = np.hstack(
        [
            matrix[2, 1] - matrix[1, 2],
            matrix[0, 2] - matrix[2, 0],
            matrix[1, 0] - matrix[0, 1],
            1 + t,
        ]
    )

    quat = q0
    max_val = matrix[0, 0]

    quat = np.where(matrix[1, 1] >= max_val, q1, quat)
    max_val = np.where(matrix[1, 1] >= max_val, matrix[1, 1], max_val)

    quat = np.where(matrix[2, 2] >= max_val, q2, quat)
    max_val = np.where(matrix[2, 2] >= max_val, matrix[2, 2], max_val)

    quat = np.where(t >= max_val, q3, quat)
    quat = quat / np.linalg.norm(quat)

    quat = np.roll(quat, 1)  # Convert to scalar-first format
    return quat


def euler_to_quaternion(angles: np.ndarray, seq: str = "xyz") -> np.ndarray:
    """Convert Euler angles in radians to unit quaternion.

    This method uses the same notation and conventions as the SciPy Rotation class.
    See the SciPy documentation for more details.  Some common examples:

    - 'xyz': Extrinsic rotations about x, then y, then z axes (classical roll,
        pitch, yaw sequence)
    - 'ZXZ': Rotation from perifocal (Ω, i, ω) frame (right ascension of ascending
        node, inclination, argument of perigee) used by Kepler orbital elements to
        ECI (Earth-Centered Inertial) frame

    Parameters
    ----------
    angles : array_like, shape (N,) or (1, N) or (N, 1)
        Euler angles in radians. The number of angles must match the length of `seq`.
    seq : str
        Specifies sequence of axes for rotations. Up to 3 characters belonging to
        the set {'x', 'y', 'z'} or {'X', 'Y', 'Z'}. Lowercase characters
        correspond to extrinsic rotations about the fixed axes, while uppercase
        characters correspond to intrinsic rotations about the rotating axes.
        Examples include 'xyz', 'ZYX', 'xzx', etc.

    Returns
    -------
    np.ndarray, shape (4,)
        Unit quaternion [q0, q1, q2, q3] corresponding to the Euler angles,
        where q0 is the scalar part.

    Raises
    ------
    ValueError
        If `seq` is not a valid sequence of axes, or if the shape of `angles` does
        not match the length of `seq`.
    """
    intrinsic = _check_seq(seq)
    angles = _check_angles(angles, seq)

    seq = seq.lower()
    angles = angles.flatten()

    quat = _elementary_quat_compose(seq, angles, intrinsic=intrinsic)
    return cast(np.ndarray, quat / np.linalg.norm(quat))


def quaternion_to_dcm(quat: np.ndarray) -> np.ndarray:
    """Direction cosine matrix from unit quaternion

    If the quaternion represents the attitude of a body B relative to a frame A,
    then this function returns the matrix R_BA that transforms vectors from
    frame A to frame B.  Specifically, for a vector v_A expressed in frame A,
    the corresponding vector in frame B is given by ``v_B = R_BA @ v_A``.

    The inverse transformation can be obtained by transposing this matrix:
    ``R_AB = R_BA.T``.

    Note that this function assumes the scalar-first convention, and assumes the
    quaternion is normalized.

    Parameters
    ----------
    quat : array_like, shape (4,)
        Unit quaternion representing rotation from frame A to frame B,
        in the format [w, x, y, z] where w is the scalar part.

    Returns
    -------
    np.ndarray, shape (3, 3)
        Direction cosine matrix R_BA that transforms vectors from frame A to frame B.
    """
    w, x, y, z = quat
    x2 = x * x
    y2 = y * y
    z2 = z * z
    w2 = w * w
    xy = x * y
    xz = x * z
    xw = x * w
    yz = y * z
    yw = y * w
    zw = z * w

    return np.array(
        [
            [w2 + x2 - y2 - z2, 2 * (xy - zw), 2 * (xz + yw)],
            [2 * (xy + zw), w2 - x2 + y2 - z2, 2 * (yz - xw)],
            [2 * (xz - yw), 2 * (yz + xw), w2 - x2 - y2 + z2],
        ],
        like=quat,
    ).T


# See: https://github.com/scipy/scipy/blob/3ead2b543df7c7c78619e20f0cb6139e344a8866/scipy/spatial/transform/_rotation_cy.pyx#L774-L851  # ruff: noqa: E501
def quaternion_to_euler(q: np.ndarray, seq: str = "xyz") -> np.ndarray:
    """Convert unit quaternion to roll-pitch-yaw Euler angles.

    This method uses the same notation and conventions as the SciPy Rotation class.
    See the SciPy documentation and :py:func:`euler_to_quaternion` for more details.

    The algorithm is based on the method described in [1]_ as implemented in SciPy.

    Parameters
    ----------
    q : array_like, shape (4,)
        Unit quaternion representing rotation, in the format [q0, q1, q2, q3]
        where q0 is the scalar part.
    seq : str, optional
        Sequence of axes for Euler angles (default is 'xyz').

    Returns
    -------
    np.ndarray, shape (3,)
        Euler angle sequence corresponding to the quaternion.

    References
    ----------
    .. [1] Bernardes E, Viollet S (2022) Quaternion to Euler angles
            conversion: A direct, general and computationally efficient
            method. PLoS ONE 17(11): e0276302.
            https://doi.org/10.1371/journal.pone.0276302
    """
    if len(seq) != 3:
        raise ValueError("Expected `seq` to be a string of 3 characters")

    intrinsic = _check_seq(seq)
    seq = seq.lower()

    if intrinsic:
        seq = seq[::-1]

    # Note: the sequence is "static" from a symbolic computation point of view,
    # meaning that the indices are known at "compile-time" and all logic on indices
    # will be evaluated in standard Python.
    i, j, k = (_elementary_basis_index(axis) for axis in seq)

    symmetric = i == k
    if symmetric:
        k = 6 - i - j

    # 0. Check if permutation is odd or even
    sign = (i - j) * (j - k) * (k - i) // 2

    # 1. Permute quaternion components
    if symmetric:
        a, b, c, d = (q[0], q[i], q[j], q[k] * sign)
    else:
        a, b, c, d = (
            q[0] - q[j],
            q[i] + q[k] * sign,
            q[j] + q[0],
            q[k] * sign - q[i],
        )

    # 2. Compute second angle
    angles = np.zeros(3, like=q)
    angles[1] = 2 * np.arctan2(np.hypot(c, d), np.hypot(a, b))

    # 3. Compute first and third angles
    half_sum = np.arctan2(b, a)
    half_diff = np.arctan2(d, c)

    angles[0] = half_sum - half_diff
    angles[2] = half_sum + half_diff

    # Handle singularities
    s_zero = abs(angles[1]) <= 1e-7
    s_pi = abs(angles[1] - np.pi) <= 1e-7

    angles[0] = np.where(s_zero, 2 * half_sum, angles[0])
    angles[2] = np.where(s_zero, 0.0, angles[2])

    angles[0] = np.where(s_pi, -2 * half_diff, angles[0])
    angles[2] = np.where(s_pi, 0.0, angles[2])

    # Tait-Bryan/asymmetric sequences
    if not symmetric:
        angles[2] *= sign
        angles[1] -= np.pi / 2

    if intrinsic:
        angles = angles[::-1]

    angles = (angles + np.pi) % (2 * np.pi) - np.pi

    return angles


def quaternion_kinematics(
    q: np.ndarray, w: np.ndarray, baumgarte: float | None = None
) -> np.ndarray:
    """Quaternion kinematical equations

    If the rotation represents the attitude of a body B relative to a
    frame A, then w should be the body relative angular velocity, i.e. ω_B.

    The derivative is computed using quaternion kinematics:
        dq/dt = 0.5 * q ⊗ [0, ω]
    where ⊗ is the quaternion multiplication operator.

    The method optionally support Baumgarte stabilization to preserve
    unit normalization.  For a stabilization factor λ, the full
    time derivative is:
        dq/dt = 0.5 * q ⊗ [0, ω] - λ * (||q||² - 1) * q

    Parameters
    ----------
    q : array_like, shape (4,)
        Unit quaternion representing rotation, in the format [q0, q1, q2, q3]
        where q0 is the scalar part.
    w : array_like, shape (3,)
        Angular velocity vector in body frame.
    baumgarte : float, optional
        Baumgarte stabilization factor. If not None, Baumgarte stabilization is
        applied to enforce unit norm constraint. Default is None (no stabilization).

    Returns
    -------
    np.ndarray, shape (4,)
        Time derivative of the quaternion.
    """
    omega = np.array([0, *w], like=q)
    q_dot = 0.5 * quaternion_multiply(q, omega)

    # Baumgarte stabilization to enforce unit norm constraint
    if baumgarte is not None:
        q_dot -= baumgarte * (np.dot(q, q) - 1) * q

    return q_dot
